#!/usr/bin/env pytest

from ornithology import *
import htcondor2
import os
import pytest
from time import time as now
from time import sleep
from pathlib import Path
from shutil import rmtree
from typing import Callable, Optional, Any

#-----------------------------------------------------------------------------------------
class SubmissionError(Exception):
        """Custom exception to raise in job submission helper functions upon failure."""
        def __init__(self, message="Failed to submit job(s)"):
            self.message = message
            super().__init__(self.message)

#-----------------------------------------------------------------------------------------
TEST_CASE_ATTR = "TestCase"
TEST_CASE_ENV = "TEST_CASE"

FILENAME_SUB = "sleep.sub"
FILENAME_DAG = "test.dag"
FILENAME_JOBSET = "job.set"

JOB_LOG = "job.log"

KEY_VAL = "value"
KEY_OVERRIDE = "override"
KEY_NOUN = "noun"

#-----------------------------------------------------------------------------------------
@standup
def the_condor(test_dir):
    dagman_append = test_dir / f".{FILENAME_DAG}.append"
    with open(dagman_append, "w") as f:
        f.write(f"""
submit_event_user_notes = $ENV({TEST_CASE_ENV})
My.{TEST_CASE_ATTR} = "$ENV({TEST_CASE_ENV})"
""")

    with Condor(test_dir / "the_condor", config={"USE_JOBSETS": True, "DAGMAN_INSERT_SUB_FILE": dagman_append}) as condor:
        yield condor

#-----------------------------------------------------------------------------------------
def write_submit(sleep: Path) -> Path:
    path = Path(FILENAME_SUB)
    with open(path, "w") as f:
        f.write(f"""# Simple sleep job
executable = {sleep}
arguments  = 0
log = {JOB_LOG}

submit_event_user_notes = $ENV({TEST_CASE_ENV})

My.{TEST_CASE_ATTR} = "$ENV({TEST_CASE_ENV})"

queue
""")

    return path

def write_dag(sleep: Path, sleep_submit: Path) -> Path:
    path = Path(FILENAME_DAG)
    with open(path, "w") as f:
        f.write(f"""# Simple DAG file
ENV GET {TEST_CASE_ENV}

SUBMIT-DESCRIPTION sleep @=desc
    executable = {sleep}
    arguments  = 0
    submit_event_user_notes = $ENV({TEST_CASE_ENV})
    My.{TEST_CASE_ATTR} = "$ENV({TEST_CASE_ENV})"
    queue
@desc

JOB DESC sleep
JOB INLINE @=desc
    executable = {sleep}
    arguments  = 0
    submit_event_user_notes = $ENV({TEST_CASE_ENV})
    My.{TEST_CASE_ATTR} = "$ENV({TEST_CASE_ENV})"
    queue
@desc
JOB EXTERNAL {sleep_submit}
""")

    return path

def prepare_job(sleep: Path) -> Path:
    return write_submit(sleep)

def prepare_dag(sleep: Path) -> Path:
    sub = write_submit(sleep)
    return write_dag(sleep, sub)

def prepare_jobset(sleep: Path) -> Path:
    path = Path(FILENAME_JOBSET)
    with open(path, "w") as f:
        f.write(f"""
name = JobSubmitMethodTest

iterator = table var {{
     Job1
     Job2
}}

job {{
     executable = {sleep}
     arguments  = 0
     log        = {JOB_LOG}

     submit_event_user_notes = $ENV({TEST_CASE_ENV})
     My.{TEST_CASE_ATTR} = "$ENV({TEST_CASE_ENV})"

     queue
}}
""")

    return path

#-----------------------------------------------------------------------------------------
def check_undefined(condor: Condor, test: Path, **kwargs) -> None:
    try:
        with condor.use_config():
            with open(test, "r") as f:
                desc = htcondor2.Submit(f.read())
            desc.setSubmitMethod(-1, True)
            schedd = htcondor2.Schedd()
            result = schedd.submit(desc)
    except Exception as e:
        raise SubmissionError(f"Failed to submit via python API: {e}")

def check_condor_submit(condor: Condor, test: Path, **kwargs) -> None:
    p = condor.run_command(["condor_submit", test])
    if p.returncode != 0:
        raise SubmissionError(f"Failed to submit via condor_submit:\n{p.stdout}\n{p.stderr}")

def check_submit_dag_direct(condor: Condor, test: Path, **kwargs) -> None:
    p = condor.run_command(["condor_submit_dag", "-SubmitMethod", "1", test])
    if p.returncode != 0:
        raise SubmissionError(f"Failed to submit via condor_dag_submit:\n{p.stdout}\n{p.stderr}")

def check_submit_dag_shell(condor: Condor, test: Path, **kwargs) -> None:
    p = condor.run_command(["condor_submit_dag", "-SubmitMethod", "0", test])
    if p.returncode != 0:
        raise SubmissionError(f"Failed to submit via condor_dag_submit:\n{p.stdout}\n{p.stderr}")

def check_dag_py_submit(condor: Condor, test: Path, **kwargs) -> None:
    try:
        with condor.use_config():
            desc = htcondor2.Submit.from_dag(str(test))
            schedd = htcondor2.Schedd()
            result = schedd.submit(desc)
    except Exception as e:
        raise SubmissionError(f"Failed to submit via python API: {e}")

def check_py_submit(condor: Condor, test: Path, **kwargs) -> str:
    ret = "UhOh..."

    try:
        with condor.use_config():
            with open(test, "r") as f:
                desc = htcondor2.Submit(f.read())
            try:
                val = kwargs.get(KEY_VAL)
                override = kwargs.get(KEY_OVERRIDE, False)
                if val is not None:
                    desc.setSubmitMethod(val, override)
                ret = str(desc.getSubmitMethod())
            except ValueError as e:
                ret = str(e)

            schedd = htcondor2.Schedd()
            result = schedd.submit(desc)
    except Exception as e:
        raise SubmissionError(f"Failed to submit via python API: {e}")

    return ret

def check_htc_submit(condor: Condor, test: Path, **kwargs) -> None:
    noun = kwargs.get(KEY_NOUN, "NO-NOUN")
    p = condor.run_command(["htcondor", noun, "submit", test])
    if p.returncode != 0:
        raise SubmissionError(f"Failed to submit via htcondor {noun} submit:\n{p.stdout}\n{p.stderr}")

#-----------------------------------------------------------------------------------------
class Details():
    def __init__(self, prep: Callable[[Path], Path], exe: Callable[[Condor, str], Optional[str]], result: dict, extra: dict = {}, out: Optional[str] = None, is_dag: bool = False):
        self.prepare = prep
        self.execute = exe
        self.expected = result
        self.log = f"{FILENAME_DAG}.dagman.log" if is_dag else JOB_LOG
        self.check = out
        self.details = extra

TEST_CASES = {
    "UNDEFINED": Details(prepare_job, check_undefined, {None:1}),
    "CONDOR_SUBMIT": Details(prepare_job, check_condor_submit, {0:1}),
    "DAGMAN_DIRECT": Details(prepare_dag, check_submit_dag_direct, {0:1, 1:3}, is_dag=True),
    "DAGMAN_SHELL": Details(prepare_dag, check_submit_dag_shell, {0:4}, is_dag=True),
    "PY_DAG": Details(prepare_dag, check_dag_py_submit, {1:3, 2:1}, is_dag=True),
    "PY_DEFAULT": Details(prepare_job, check_py_submit, {2:1}, out="2"),
    "PY_INVALID": Details(prepare_job, check_py_submit, {2:1}, extra={KEY_VAL:69}, out="Submit method value must be 100 or greater, or allowed_reserved_values must be True."),
    "PY_OVERRIDE": Details(prepare_job, check_py_submit, {69:1}, extra={KEY_VAL:69, KEY_OVERRIDE:True}, out="69"),
    "PY_SET_100": Details(prepare_job, check_py_submit, {100:1}, extra={KEY_VAL:100}, out="100"),
    "PY_SET": Details(prepare_job, check_py_submit, {666:1}, extra={KEY_VAL:666}, out="666"),
    "HTC_JOB_SUBMIT": Details(prepare_job, check_htc_submit, {3:1}, extra={KEY_NOUN:"job"}),
    "HTC_DAG_SUBMIT": Details(prepare_dag, check_htc_submit, {4:1, 1:3}, extra={KEY_NOUN:"dag"}, is_dag=True),
    "HTC_JOBSET_SUBMIT": Details(prepare_jobset, check_htc_submit, {5:2}, extra={KEY_NOUN:"jobset"}),
}

#-----------------------------------------------------------------------------------------
@action
def run_dags(the_condor, test_dir, path_to_sleep):
    TESTS = dict()

    for TEST, SETUP in TEST_CASES.items():
        case_dir = test_dir / TEST
        if case_dir.exists():
            rmtree(case_dir)

        os.mkdir(case_dir)

        with ChangeDir(TEST):
            try:
                test_file = SETUP.prepare(path_to_sleep)
                os.environ[TEST_CASE_ENV] = TEST
                output = SETUP.execute(the_condor, test_file, **SETUP.details)
                del os.environ[TEST_CASE_ENV]

                TESTS[TEST] = (SETUP.log, SETUP.expected, SETUP.check, output)
            except SubmissionError as e:
                TESTS[TEST] = e

    yield TESTS

#-----------------------------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def test_info(request, run_dags):
    details = run_dags[request.param]

    if isinstance(details, SubmissionError):
        pytest.fail(str(details))

    assert isinstance(details, tuple)

    return (
        request.param,    # Test name
        details[0],       # Test log to follow
        details[1],       # Test expected JobSubmitMethod values -> Count
        details[2],       # Test expected output (None == skip)
        details[3],       # Test output
    )

@action
def test_case(test_info):
    return test_info[0]

@action
def test_log(test_info):
    return test_info[1]

@action
def test_expected(test_info):
    return test_info[2]

@action
def test_check(test_info):
    return test_info[3]

@action
def test_output(test_info):
    return test_info[4]

#-----------------------------------------------------------------------------------------
@action
def test_wait(test_case, test_log) -> None:
    with ChangeDir(test_case):
        jel = htcondor2.JobEventLog(test_log)
        start = now()
        counts = [0, 0]
        while True:
            assert now() - start <= 90, f"ERROR: Test case {test_case} job(s) timed out"
            for event in jel.events(stop_after=0):
                if event.type == htcondor2.JobEventType.SUBMIT:
                    assert event.get("LogNotes") == test_case
                    counts[0] += 1
                elif event.type == htcondor2.JobEventType.JOB_TERMINATED:
                    assert event.get("ReturnValue") == 0
                    counts[1] += 1
                    if len(set(counts)) == 1:
                        return

@action
def test_get_counts(the_condor, test_case, test_wait, test_expected) -> dict:
    schedd = the_condor.get_local_schedd()

    expected_num_ads = sum(test_expected.values())
    attempts = 0
    history = []

    while len(history) != expected_num_ads:
        if attempts > 0:
            sleep(2)

        attempts += 1
        assert attempts <= 10, f"ERROR: Failed to get history all {expected_num_ads} records for test case {test_case}"
        history = schedd.history(
            constraint=f'TestCase=="{test_case}"',
            projection=["JobSubmitMethod"],
        )

    results = dict()
    for ad in history:
        jsm = ad.get("JobSubmitMethod")
        if jsm in results:
            results[jsm] += 1
        else:
            results.update({jsm : 1})

    return results

#=========================================================================================
class TestJobSubmitMethod:
    def test_check_value_counts(self, test_get_counts, test_expected):
        assert test_get_counts == test_expected

    def test_check_output(self, test_check, test_output):
        if test_check is None:
            pytest.skip("Test case does not check output")
        assert test_check == test_output



#!/usr/bin/env pytest

import queue
import socket
import ssl
import struct
import subprocess
import threading
import time
from contextlib import closing
from pathlib import Path
import shutil
import tempfile

from ornithology import action, standup, Condor

FORWARD_ID = "http_api_forward"


class DomainSocketHTTPServer:
    def __init__(self, socket_path: Path, ssl_context: ssl.SSLContext):
        self.socket_path = Path(socket_path)
        self.ssl_context = ssl_context
        self._listener = None
        self._thread = None
        self._stop = threading.Event()
        self._ready = threading.Event()
        self.seen_requests = queue.Queue()

    def _log(self, msg: str):
        print(f"[domain-socket-server] {msg}", flush=True)

    def start(self):
        if self.socket_path.exists():
            self.socket_path.unlink()

        self.socket_path.parent.mkdir(parents=True, exist_ok=True)

        self._listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._listener.bind(str(self.socket_path))
        self._listener.listen(5)
        self._listener.settimeout(0.5)

        self._thread = threading.Thread(target=self._serve, daemon=True)
        self._thread.start()
        self._ready.wait(timeout=5)

    def stop(self):
        self._stop.set()
        if self._listener:
            try:
                self._listener.shutdown(socket.SHUT_RDWR)
            except OSError:
                # Listener might be closed or not connected
                pass
            self._listener.close()
        if self._thread:
            self._thread.join(timeout=5)
        if self.socket_path.exists():
            try:
                self.socket_path.unlink()
            except OSError:
                # File might not exist or be removed by another process
                pass

    def _serve(self):
        self._ready.set()
        while not self._stop.is_set():
            try:
                conn, _ = self._listener.accept()
            except socket.timeout:
                continue
            except OSError:
                # Listener socket closed
                break

            try:
                self._handle_domain_connection(conn)
            except Exception as e:
                self._log(f"exception while handling domain connection: {e}")
                conn.close()
                continue

    def _handle_domain_connection(self, conn: socket.socket):
        deadline = time.monotonic() + 15

        while True:
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                self._log("timeout waiting for fd arrival")
                conn.close()
                return

            conn.settimeout(remaining)
            try:
                msg, ancdata, _flags, _addr = conn.recvmsg(1024, socket.CMSG_SPACE(struct.calcsize("i")))
            except (socket.timeout, BlockingIOError, OSError):
                self._log("timeout or recvmsg error before fd arrival")
                conn.close()
                return

            if msg == b"":
                self._log("domain socket closed before fd arrival")
                conn.close()
                return

            if not ancdata:
                continue
            break

        fd = struct.unpack("i", ancdata[0][2])[0]
        client_sock = socket.socket(fileno=fd)
        client_sock.settimeout(5)

        try:
            first = client_sock.recv(1, socket.MSG_PEEK)
        except (socket.timeout, OSError):
            self._log("peek timeout/err on forwarded socket")
            client_sock.close()
            conn.close()
            return

        if first.startswith(b"\x16"):
            handler = self._handle_tls_client
            self._log("handling TLS client")
        else:
            handler = self._handle_http_client
            self._log("handling HTTP client")

        handler(client_sock)
        conn.close()

    def _http_response(self, body: bytes) -> bytes:
        return b"HTTP/1.1 200 OK\r\nContent-Length: %d\r\nConnection: close\r\n\r\n%s" % (
            len(body),
            body,
        )

    def _handle_http_client(self, client_sock: socket.socket):
        try:
            data = client_sock.recv(4096)
            self.seen_requests.put(("http", data))
            client_sock.sendall(self._http_response(b"http-ok"))
        finally:
            client_sock.close()

    def _handle_tls_client(self, client_sock: socket.socket):
        try:
            tls_sock = self.ssl_context.wrap_socket(client_sock, server_side=True)
            data = tls_sock.recv(4096)
            self.seen_requests.put(("https", data))
            tls_sock.sendall(self._http_response(b"https-ok"))
            tls_sock.close()
        finally:
            client_sock.close()


def _pick_ephemeral_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


def _create_self_signed_cert(target_dir: Path):
    key_path = target_dir / "server.key"
    cert_path = target_dir / "server.crt"

    subprocess.run(
        [
            "openssl",
            "req",
            "-new",
            "-newkey",
            "rsa:2048",
            "-nodes",
            "-x509",
            "-days",
            "1",
            "-subj",
            "/CN=localhost",
            "-keyout",
            key_path,
            "-out",
            cert_path,
        ],
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    return cert_path, key_path


@action
def shared_port_socket_dir():
    path = Path(tempfile.mkdtemp(prefix="htcondor-shared-port-", dir="/tmp"))
    yield path
    shutil.rmtree(path, ignore_errors=True)


@standup
def condor(test_dir, shared_port_socket_dir):
    shared_port_socket_dir.mkdir(parents=True, exist_ok=True)

    shared_port_port = _pick_ephemeral_port()

    config = {
        "DAEMON_LIST": "MASTER SHARED_PORT",
        "SHARED_PORT_PORT": str(shared_port_port),
        "USE_SHARED_PORT": True,
        "DAEMON_SOCKET_DIR": shared_port_socket_dir.as_posix(),
        "COLLECTOR_USES_SHARED_PORT": False,
        "SHARED_PORT_HTTP_FORWARDING_ID": FORWARD_ID,
        "SHARED_PORT_LOG": "$(LOG)/SharedPortLog",
    }

    with Condor(local_dir=test_dir / "condor", config=config) as c:
        yield c


@action
def ssl_context(test_dir):
    cert_path, key_path = _create_self_signed_cert(test_dir)
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    context.load_cert_chain(certfile=cert_path, keyfile=key_path)
    return context


@action
def domain_socket_server(shared_port_socket_dir, ssl_context):
    socket_path = shared_port_socket_dir / FORWARD_ID
    server = DomainSocketHTTPServer(socket_path, ssl_context)
    server.start()
    yield server
    server.stop()


@action
def shared_port_port(condor):
    return int(condor.run_command(["condor_config_val", "SHARED_PORT_PORT"], echo=False, suppress=True).stdout)


class TestSharedPortHTTPForwarding:
    def _read_response(self, conn: socket.socket) -> bytes:
        chunks = []
        while True:
            chunk = conn.recv(4096)
            if not chunk:
                break
            chunks.append(chunk)
        return b"".join(chunks)

    def test_http_is_forwarded(self, shared_port_port, domain_socket_server):
        with socket.create_connection(("127.0.0.1", shared_port_port), timeout=5) as conn:
            conn.sendall(b"GET / HTTP/1.1\r\nHost: localhost\r\n\r\n")
            response = self._read_response(conn)

        assert b"HTTP/1.1 200 OK" in response
        assert b"http-ok" in response

    def test_https_is_forwarded(self, shared_port_port, domain_socket_server, ssl_context):
        client_ctx = ssl.create_default_context()
        client_ctx.check_hostname = False
        client_ctx.verify_mode = ssl.CERT_NONE

        with client_ctx.wrap_socket(socket.create_connection(("127.0.0.1", shared_port_port), timeout=5), server_hostname="localhost") as conn:
            conn.sendall(b"GET /secure HTTP/1.1\r\nHost: localhost\r\n\r\n")
            response = self._read_response(conn)

        assert b"HTTP/1.1 200 OK" in response
        assert b"https-ok" in response



#!/usr/bin/env pytest

from ornithology import (
    action,
    ClusterState,
    Condor,
)

import socket

from pytest_httpserver import HTTPServer
from werkzeug.wrappers import Response

import time


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "ENABLE_URL_TRANSFERS": "TRUE",
            "FILETRANSFER_PLUGINS": "$(LIBEXEC)/curl_plugin",
        },
    ) as condor:
        yield condor


TEST_CASES = {
    "CURLE_URL_MALFORMAT": {
        "expected": {
            "ErrorType":        "Parameter",
            "PluginLaunched":   True,
            "PluginVersion":    "0.2",
        },
        "url":    "file://$(staging/match_metacompare/xac_plots.pdf",
    },
    # Let's not even and say we didn't.
    # CURLE_COULDNT_RESOLVE_PROXY
    "CURLE_COULDNT_RESOLVE_HOST": {
        "expected": {
            "ErrorType":        "Resolution",
            "FailedName":       "no-such.tld",
            "FailureType":      "Unknown",
        },
        "url":    "http://no-such.tld/file",
    },
    # I don't know how to generate this one, actually.
    # CURLE_REMOTE_ACCESS_DENIED
    "CURLE_FILE_COULDNT_READ_FILE": {
        "expected": {
            "ErrorType":        "Specification",
        },
        "url":    "file:///no-such.file",
    },
    "HTTP_400": {
        "expected": {
            "ErrorType":        "Specification",
            "FailedServer":     "localhost:{open_port}",
        },
        "url":    "http://localhost:{open_port}/400",
    },
    "HTTP_403": {
        "expected": {
            "ErrorType":        "Authorization",
            "FailureType":      "Authorization",
            "FailedServer":     "localhost:{open_port}",
        },
        "url":    "http://localhost:{open_port}/403",
    },
    "HTTP_404": {
        "expected": {
            "ErrorType":        "Specification",
            "FailedServer":     "localhost:{open_port}",
        },
        "url":    "http://localhost:{open_port}/404",
    },
    "CURLE_ABORTED_BY_CALLBACK": {
        "expected": {
            "ErrorType":        "Transfer",
            "FailedServer":     "Unknown",
            "FailureType":      "TooSlow",
        },
        "url":    "http://localhost:{slow_port}/slow",
    },

#    "CURLE_TOO_MANY_REDIRECTS": {
#        "expected": {
#            "ErrorType":        "Specification",
#            "FailedServer":     "Unknown",
#        },
#        "url":    "http://localhost:{redirect_port}/tmr",
#    },

    "CURLE_COULDNT_CONNECT": {
        "expected": {
            "ErrorType":        "Contact",
            "FailedServer":     "Unknown",
        },
        "url":    "http://localhost:{closed_port}/file",
    },

    # I have no idea what triggers these.
    # CURLE_SEND_ERROR
    # CURLE_RECV_ERROR
    # Let's not worry about setting https up in just the wrong way quite yet.
    # CURLE_PEER_FAILED_VERIFICATION
    # Let's not worry about setting https up correctly just yet.
    # CURLE_SSL_CONNECT_ERROR
}


@action
def the_job_description(path_to_sleep):
    return {
        "LeaveJobInQueue":              "true",
        "executable":                   path_to_sleep,
        "transfer_executable":          "true",
        "arguments":                    "20",
        "when_to_transfer_output":      "ON_EXIT",
        "should_transfer_files":        "true",
        "environment":                  "CURL_MAX_RETRY_ATTEMPTS=1;CURL_IGNORE_HTTP_PROXY=1"
    }


@action
def the_closed_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    (address, port) = s.getsockname()
    yield port
    s.close()


def slow_reply_handler(request):
    time.sleep(32)
    return Response()


@action
def the_open_port():
    server = HTTPServer()
    if not server.is_running():
        server.start()

    # Arguably, all of these kind of lines throughout this whole test
    # should be in the corresponding test case, instead.
    server.expect_request("/400").respond_with_data("", status=400)
    server.expect_request("/403").respond_with_data("", status=403)
    server.expect_request("/404").respond_with_data("", status=404)


    yield server.port

    server.stop()


@action
def the_slow_port():
    server = HTTPServer()
    if not server.is_running():
        server.start()

    # This must be its own server, because otherwise the order of the requests
    # matter, since each server is single-threaded.
    server.expect_request("/slow").respond_with_handler(slow_reply_handler)

    yield server.port

    server.stop()


@action
def the_redirect_port():
    server = HTTPServer()
    if not server.is_running():
        server.start()

    # Joyously, curl doesn't respond to this consistently: some of the time
    # it reports error 78 (CURLE_REMOTE_FILE_NOT_FOUND) and some of the time
    # it reports error 47 (CURLE_TOO_MANY_REDIRECTS).  (This is assuming the
    # plug-in has been modified to set CURLOPT_MAXREDIRS, which doesn't have
    # a default value in older versions of libcurl.)

    server.expect_request("/tmr").respond_with_data("", status=302,
        headers=[{"Location", f"http://localhost:{server.port}/tms"}]
    )
    server.expect_request("/tms").respond_with_data("", status=302,
        headers=[{"Location", f"http://localhost:{server.port}/tmr"}]
    )

    yield server.port

    server.stop()


@action
def the_job_handles(test_dir, the_condor, the_job_description, the_closed_port, the_open_port, the_slow_port, the_redirect_port):
    job_handles = {}

    for name, test_case in TEST_CASES.items():
        complete_job_description = {
            ** the_job_description,
            "transfer_input_files": test_case["url"].format(
                closed_port=the_closed_port,
                open_port=the_open_port,
                slow_port=the_slow_port,
                redirect_port=the_redirect_port,
            ),
            "log": (test_dir / f"{name}.log").as_posix(),
        }
        job_handle = the_condor.submit(
            description = complete_job_description,
            count = 1
        )

        job_handles[name] = job_handle

    yield job_handles


@action(params={name: name for name in TEST_CASES})
def test_case_pair(request, the_job_handles):
    return (request.param, the_job_handles[request.param])


@action
def test_name(test_case_pair):
    return test_case_pair[0]


@action
def test_job(test_case_pair):
    return test_case_pair[1]


@action
def test_case(test_case_pair):
    return TEST_CASES[test_case_pair[0]]


@action
def expected_value(test_case, the_closed_port, the_open_port):
    expected_value = test_case["expected"]
    for key in expected_value.keys():
        if isinstance(expected_value[key], str):
            expected_value[key] = expected_value[key].format(
                closed_port=the_closed_port, open_port=the_open_port
            )
    return expected_value


@action
def actual_value(the_condor, test_job):
    assert test_job.wait(
        timeout=60,
        condition=ClusterState.all_held,
        fail_condition=ClusterState.any_complete,
    )

    history = the_condor.get_local_schedd().jobEpochHistory(
        f"ClusterId=={test_job.clusterid}",
        ad_type="INPUT",
    )

    return history[0]["InputPluginResultList"][0]["TransferErrorData"][0]


class TestCurlStructuredError:

    def test_epoch_log_as_expected(self, expected_value, actual_value):
        for key, value in expected_value.items():
            assert actual_value.get(key) == value



import enum

# NOTE: For internal use currently
#       Keep up to date with HistoryRecordSource
#       enum in src/condor_tools/history.cpp
class HistorySrc(enum.IntEnum):
    """
    An enumeration of history record sources.

    .. attribute:: ScheddJob

        Search final job history in :macro:`HISTORY`

    .. attribute:: StartdJob

        Search job history from Startd's :macro:`HISTORY`

    .. attribute:: JobEpoch

        Search historical records from :macro:`JOB_EPOCH_HISTORY`.
        Currently contains job epoch, transfer, and spawn records.

    .. attribute:: Daemon

        Search for historical daemon records from ``<SUBSYS>_DAEMON_HISTORY``.

        .. note::

            Requires history queries AdType to be set to the desired
            daemon subsystem.

    :meta private:
    """

    ScheddJob = 0
    StartdJob = 1
    JobEpoch = 2
    Daemon = 3


#!/usr/bin/env pytest

from pathlib import Path

from ornithology import (
    action,
    Condor,
    ClusterState,
)

import logging


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            'SCHEDD_DEBUG':                             'D_TEST',
            'BYTES_REQUIRED_TO_QUEUE_FOR_TRANSFER':     256,
        },
    ) as the_condor:
        (test_dir / 'small.txt').write_text("one two three")
        (test_dir / 'large.txt').write_text(
"""
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
one two three four five six seven eight nine ten
"""
        )
        yield the_condor


@action
def the_job_description(path_to_sleep, test_dir):
    return {
        "executable":                   path_to_sleep,
        "arguments":                    5,
        "transfer_executable":          "false",
        "should_transfer_files":        "true",
        "when_to_transfer_output":      "ON_EXIT",
        "log":                          test_dir / "test_job_$(CLUSTER).log",
    }


@action
def completed_small_job(test_dir, the_condor, the_job_description):
    complete_job_description = {
        ** the_job_description,
        ** { 'transfer_input_files': (test_dir / 'small.txt').as_posix(), },
    }

    the_job_handle = the_condor.submit(
        description=complete_job_description,
        count=1,
    )

    assert the_job_handle.wait(
        timeout=60,
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
    )

    return the_job_handle;


@action
def completed_large_job(test_dir, the_condor, the_job_description):
    complete_job_description = {
        ** the_job_description,
        ** { 'transfer_input_files': (test_dir / 'large.txt').as_posix(), },
    }

    the_job_handle = the_condor.submit(
        description=complete_job_description,
        count=1,
    )

    assert the_job_handle.wait(
        timeout=60,
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
    )

    return the_job_handle;


@action
def the_schedd_log(the_condor, completed_small_job, completed_large_job):
    return the_condor.schedd_log.open()


@action
def enqueued_jobs(the_schedd_log):
    jobs = []

    for message in the_schedd_log.read():
        if 'TransferQueueManager: enqueueing' in message:
            words = str(message).split(' ')
            for i in range(len(words)):
                if words[i] == "job":
                    jobs.append(words[i+1])

    return jobs


class TestDontQueueSmallSandboxes:

    def test_queue_skipped(self, completed_small_job, enqueued_jobs):
        jobID = f"{completed_small_job.clusterid}.0"
        assert not jobID in enqueued_jobs

    def test_queue_entered(self, completed_large_job, enqueued_jobs):
        jobID = f"{completed_large_job.clusterid}.0"
        assert jobID in enqueued_jobs



#!/usr/bin/env pytest


# Test that the job attribute InitialWaitDuration 
# is set, and to a reasonable value

import logging

from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

@standup
def condor(test_dir):
    with Condor(test_dir / "condor") as condor:
        yield condor

@action
def completed_test_job(condor):
    completed_test_job = condor.submit(
        {
            "executable":           "/bin/sleep",
            "arguments":            "0",
            "log":                  "log",
            "leave_in_queue":       "true",
        }
    )

    assert completed_test_job.wait(
        condition=ClusterState.all_terminal,
        timeout=300,
    )
    return completed_test_job

class TestWaitDuration:
    def test_job(self, completed_test_job):

        the_job_ad = completed_test_job.query()[0]
        wait_time = the_job_ad["InitialWaitDuration"]
        assert (wait_time >= 0) & (wait_time < 300)



#!/usr/bin/env pytest

# This test (test_dagman_check_q_and_exit.py) verifies that
# DAGMan will check the local schedd queue for associated jobs
# and rescue/abort if a job pending nodes jobs is not found

from ornithology import *
import htcondor2 as htcondor
import os

#--------------------------------------------------------------------------
@standup
def condor(test_dir):
    with Condor(local_dir=test_dir / "condor", config={
            "DAGMAN_USE_DIRECT_SUBMIT" : True,
            "DAGMAN_PENDING_REPORT_INTERVAL" : 3,
            "DAGMAN_CHECK_QUEUE_INTERVAL" : 3}) as condor:
        yield condor

#--------------------------------------------------------------------------
@action
def write_chmod_script():
    SCRIPT_NAME = "chmod.py"
    with open(SCRIPT_NAME, "w") as f:
        f.write(
"""#!/usr/bin/env python3
import sys
import os

os.chmod(sys.argv[1], 0o0000)
""")
    os.chmod(SCRIPT_NAME, 0o0777)
    return SCRIPT_NAME

#--------------------------------------------------------------------------
@action
def write_dag(path_to_sleep, write_chmod_script):
    DAG_FILENAME = "test.dag"
    with open(DAG_FILENAME, "w") as f:
        f.write(f"""
JOB A {{
    universe   = local
    executable = {write_chmod_script}
    arguments  = $(dagman_log)
    error      = script.err
    log        = $(JOB).log
}}

JOB B {{
    executable = {path_to_sleep}
    arguments  = 0
    log        = $(JOB).log
}}

PARENT A CHILD B
""")
    return DAG_FILENAME

#--------------------------------------------------------------------------
@action
def run_dag(condor, write_dag):
    dag = htcondor.Submit.from_dag(write_dag)
    dagman_job = condor.submit(dag)
    assert dagman_job.wait(condition=ClusterState.all_complete, timeout=45)
    return write_dag

#--------------------------------------------------------------------------
RECOVERY_DAG = "recovery.dag"
@action
def write_recovery_log():
    with open(f"{RECOVERY_DAG}.nodes.log", "w") as f:
        f.write("""000 (999.000.000) 2024-03-27 12:20:00 Job submitted from host: <11.222.33.444:52722?addrs=11.222.33.444-52722&alias=random.host.name&noUDP&sock=schedd_29282_984b>
    DAG Node: RECOVERY
...
""")

#--------------------------------------------------------------------------
@action
def write_recovery_dag(path_to_sleep, write_recovery_log):
    with open(RECOVERY_DAG, "w") as f:
        f.write(f"""
JOB RECOVERY {{
    executable = {path_to_sleep}
    arguments  = 0
}}
""")

#--------------------------------------------------------------------------
@action
def run_recovery_dag(condor, write_recovery_dag):
    assert os.path.exists(f"{RECOVERY_DAG}.nodes.log")
    dag = htcondor.Submit.from_dag(RECOVERY_DAG, {"DoRecov" : True})
    dagman_job = condor.submit(dag)
    assert dagman_job.wait(condition=ClusterState.all_complete, timeout=30)
    return RECOVERY_DAG

#==========================================================================
class TestDAGManQueryCheck:
    def check_debug(self, dag_file):
        found_errmsg = False
        with open(str(dag_file) + ".dagman.out", "r") as f:
            for line in f:
                if "ERROR: DAGMan lost track of node" in line:
                    found_errmsg = True
                    break
                elif "is in submitted state with no recorded job events!" in line:
                    found_errmsg = True
                    break
        return found_errmsg

    def test_recovery_mode_files(self, run_recovery_dag):
        assert os.path.exists(str(run_recovery_dag) + ".rescue001")
        assert self.check_debug(run_recovery_dag)

    def test_run_dag(self, run_dag):
        assert True

    def test_check_files(self, run_dag):
        rescue_file = str(run_dag) + ".rescue001"
        assert os.path.exists(rescue_file)
        jobA_done = False
        with open(str(rescue_file), "r") as f:
            for line in f:
                if line[:6] == "DONE A":
                    jobA_done = True
                    break
        assert jobA_done
        assert self.check_debug(run_dag)
        assert os.stat("B.log").st_size != 0



#!/usr/bin/env pytest

# Test that the schedd honors SYSTEM_MAX_RELEASES

from ornithology import *

#
# Setup a personal condor with just a schedd - no need for anything else
#
@standup
def condor(test_dir):
    with Condor(
        local_dir=test_dir / "condor", 
        config={
            "DAEMON_LIST": "SCHEDD MASTER",
            "USE_SHARED_PORT": False,
            "PERSONAL_CONDOR_IS_SUPER_USER": False,
            "SYSTEM_MAX_RELEASES": 3
        }
    ) as condor:
        yield condor

#
# Submit a job to be put on hold
#
@action
def job(condor, path_to_sleep):
    handle = condor.submit(
            description={"executable": path_to_sleep,
                "arguments": "3600",
                "universe": "scheduler",
                "log": "job_events.log"
                },
        count=1,
    )

    # Hold the job for the first time
    handle.hold()

    # Wait for job to go on hold the first time...
    assert handle.wait(condition=ClusterState.any_held,timeout=60)

    for iteration in range(3):
         handle.release()
         # ...and now wait for the job to be released in the event log....
         assert(handle.wait(condition=ClusterState.none_held,timeout=60))
         #
         handle.hold()
         assert(handle.wait(condition=ClusterState.all_held,timeout=60))

    assert handle.wait(condition=ClusterState.all_held,timeout=60)

    # This release should fail to release it
    handle.release()
    assert(handle.wait(condition=ClusterState.all_held,timeout=60))

    # Return the first (and only) job ad in the cluster for testing class to reference
    return handle.query()[0]

#
class TestMaxReleases:
    # Methods that begin with test_* are tests.

    def test_max_releases(self, job):
        assert job["NumHolds"] == 4



"""Generate vacate-code-combined-table.rst during Sphinx builds.

   This extension owns all logic for combining:
   1) the header row from vacate-code-table.rst,
   2) the first three columns from the data rows in hold-table.rst, and
   3) the data rows from vacate-code-table.rst.
"""

from __future__ import annotations

import os
import re
from typing import List

GENERATED_REL_PATH = "codes-other-values/vacate-code-combined-table.rst"
VACATE_TABLE_REL_PATH = "codes-other-values/vacate-code-table.rst"
HOLD_REASON_CODES_REL_PATH = "codes-other-values/hold-reason-codes.rst"


class VacateCombinedTableGenerator:
    """Create the combined vacate/hold grid table include file."""

    def __init__(self, docs_root: str) -> None:
        self.docs_root = docs_root
        self.out_path = os.path.join(docs_root, GENERATED_REL_PATH)
        self.vacate_table_path = os.path.join(docs_root, VACATE_TABLE_REL_PATH)
        hold_reason_codes_path = os.path.join(docs_root, HOLD_REASON_CODES_REL_PATH)
        self.hold_table_path = self._resolve_hold_table_path(hold_reason_codes_path)

    def _resolve_hold_table_path(self, hold_reason_codes_path: str) -> str:
        with open(hold_reason_codes_path, encoding="utf-8") as f:
            hold_reason_codes_text = f.read()

        include_match = re.search(
            r"^\s*\.\.\s+include::\s+(/codes-other-values/[^\s]+)\s*$",
            hold_reason_codes_text,
            re.MULTILINE,
        )
        if include_match:
            hold_table_rel = include_match.group(1).lstrip("/")
            return os.path.join(self.docs_root, hold_table_rel)

        return hold_reason_codes_path

    @staticmethod
    def _extract_grid_table_lines(text: str) -> List[str]:
        lines = []
        for line in text.splitlines():
            stripped = line.lstrip()
            if stripped.startswith("+") or stripped.startswith("|"):
                lines.append(stripped.rstrip())
        return lines

    @staticmethod
    def _split_grid_header_and_body(table_lines: List[str]) -> tuple[List[str], List[str]]:
        for i, line in enumerate(table_lines):
            if "=" in line and line.lstrip().startswith("+"):
                return table_lines[: i + 1], table_lines[i + 1 :]
        return table_lines, []

    @staticmethod
    def _parse_grid_widths(border_line: str) -> List[int]:
        parts = border_line.split("+")[1:-1]
        return [len(p) for p in parts]

    @staticmethod
    def _parse_grid_boundaries(border_line: str) -> List[int]:
        return [i for i, ch in enumerate(border_line) if ch == "+"]

    @staticmethod
    def _parse_grid_rows(body_lines: List[str], boundaries: List[int]) -> List[List[str]]:
        row_blocks = []
        current = []

        for line in body_lines:
            if line.startswith("|"):
                current.append(line.rstrip("\n"))
                continue

            if line.startswith("+") and current:
                row_blocks.append(current)
                current = []

        rows = []
        for block in row_blocks:
            n_cols = len(boundaries) - 1
            cols = [[] for _ in range(n_cols)]
            for content_line in block:
                for i in range(n_cols):
                    start = boundaries[i] + 1
                    end = boundaries[i + 1]
                    val = content_line[start:end] if end <= len(content_line) else ""
                    cols[i].append(val.strip())
            rows.append(["\n".join(col).rstrip() for col in cols])

        return rows

    @staticmethod
    def _render_grid_separator(indent: str, widths: List[int], fill: str = "-") -> str:
        return indent + "+" + "+".join([fill * w for w in widths]) + "+"

    @staticmethod
    def _render_grid_row(indent: str, widths: List[int], cells: List[str]) -> List[str]:
        normalized = []
        for c in cells:
            parts = c.split("\n") if c else [""]
            normalized.append(parts)

        out = []
        max_lines = max(len(c) for c in normalized)
        for i in range(max_lines):
            line = indent + "|"
            for col, width in enumerate(widths):
                text = normalized[col][i] if i < len(normalized[col]) else ""
                line += " " + text.ljust(width - 1) + "|"
            out.append(line)
        return out

    def generate(self) -> None:
        with open(self.vacate_table_path, encoding="utf-8") as f:
            vacate_text = f.read()
        with open(self.hold_table_path, encoding="utf-8") as f:
            hold_text = f.read()

        vacate_lines = self._extract_grid_table_lines(vacate_text)
        hold_lines = self._extract_grid_table_lines(hold_text)

        vacate_header, vacate_body = self._split_grid_header_and_body(vacate_lines)
        _, hold_body = self._split_grid_header_and_body(hold_lines)

        indent = vacate_header[0][: len(vacate_header[0]) - len(vacate_header[0].lstrip())]
        widths = self._parse_grid_widths(vacate_header[0])
        vacate_boundaries = self._parse_grid_boundaries(vacate_header[0])
        hold_boundaries = self._parse_grid_boundaries(hold_lines[0])

        hold_rows = self._parse_grid_rows(hold_body, hold_boundaries)
        hold_rows_3col = [row[:3] for row in hold_rows]
        vacate_rows = self._parse_grid_rows(vacate_body, vacate_boundaries)

        combined = list(vacate_header)
        for row in hold_rows_3col:
            combined.extend(self._render_grid_row(indent, widths, row))
            combined.append(self._render_grid_separator(indent, widths, "-"))
        for row in vacate_rows:
            combined.extend(self._render_grid_row(indent, widths, row))
            combined.append(self._render_grid_separator(indent, widths, "-"))

        content = (
            ".. This file is generated at build time by docs/extensions/vacate_combined_table.py.\n"
            ".. Do not edit manually.\n\n"
            ".. This file contains the complete Vacate Reason Codes table auto-generated by combining:\n"
            ".. 1) the header row from vacate-code-table.rst,\n"
            ".. 2) the first three columns from the data rows in hold-table.rst, and\n"
            ".. 3) the data rows from vacate-code-table.rst.\n\n"
            + "\n".join(combined)
            + "\n"
        )

        with open(self.out_path, "w", encoding="utf-8") as f:
            f.write(content)


def _on_builder_inited(app) -> None:
    generator = VacateCombinedTableGenerator(app.confdir)
    generator.generate()


def _on_config_inited(app, config) -> None:
    if GENERATED_REL_PATH not in config.exclude_patterns:
        config.exclude_patterns.append(GENERATED_REL_PATH)


def setup(app):
    app.connect("config-inited", _on_config_inited)
    app.connect("builder-inited", _on_builder_inited)
    return {
        "version": "1.0",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



#!/usr/bin/env python3

############################################################################
## This debug plugin takes URL's encoded with how to function via a main
## command and potentially encoded information about what ClassAd information
## to return. URL information encoding is only applicable to two commands
## (see below). All encoding is specified via a keyword option, an option
## source type, and the encoding details wrapped in square brackets.
##
## Available Commands:
##     - SLEEP     -> sleep for specified time
##     - SIGNAL    -> Send specified signal to self  (as defined by python signal module)
##     - EXIT      -> Exit immediately with specified code
##     - EXITNOAD  -> Exit immediately with specified code without producing a result ClassAd (i.e. for simulating a plugin failure that doesn't produce a result ad)
##     - EXITBADAD -> Exit immediately with specified code producing an unparseable ClassAd (i.e. for simulating EP disk filling failure where the plugin attempts to write an ad but fails to do so correctly)
##     - ERROR     -> Mimic a transfer failure using information encoded in
##                    URL to create returned ClassAd
##     - SUCCESS   -> As ERROR but successful transfer
##     - ECHO      -> Specify message to print to log
##     - VERBOSITY -> Set logging level inline (allows dynamic verbosity changing for specific URLs)
##
## Available option encodings (excluding delete option):
##     - None: option -> ../resolution/..
##     - Inline: option[ClassAd] -> ../resolution[foo="bar";baz=false;DeveloperData=[A=1;B=2.45;];]/..
##     - File: option~[path/filename] -> ../resolution~[/home/miron/test.ad]/..
##     - Script: option#[./executable] -> ../resolution#[./generate_dynamic_ad.py]/..
##
## Encoding options as detailed via: https://docs.google.com/document/d/1Gbb7XDOtmLX_xcfmCwP8KAZ_s6PDDGgCamGoDl145Ww/edit?usp=sharing
##     - general        -> Update full result ad with encoded ClassAd
##     - delete         -> Delete listed attributes from full general info to add to result ad (delete[foo,bar,...])
##     - parameter      -> Add parameter failure attempt to TransferErrorData
##     - resolution     -> Add resolution failure attempt to TransferErrorData
##     - contact        -> Add contact failure attempt to TransferErrorData
##     - authorization  -> Add authorization failure attempt to TransferErrorData
##     - specification  -> Add specification failure attempt to TransferErrorData
##     - transfer       -> Add transfer failure attempt to TransferErrorData
############################################################################

import sys
import os
import argparse
import textwrap
import time
import signal
import re
import enum
import subprocess
import datetime
from typing import Union, Tuple

# ----------------------------------------------------------------------------
EXIT_SUCCESS = 0
EXIT_SETUP_FAILURE = 125
EXIT_INVALID_URL = 124

# ----------------------------------------------------------------------------
try:
    import classad2 as classad
except ImportError:
    sys.stderr.write("Failed to import HTCondor classad2 module\n")
    sys.exit(EXIT_SETUP_FAILURE)


# ----------------------------------------------------------------------------
class DebugUrlParseError(Exception):
    """Custom URL parse failure exception"""

    def __init__(self, msg=None):
        self.message = msg if msg is not None else "Failed to parse debug plugin encoded URL"
        super().__init__(self.message)


# ----------------------------------------------------------------------------
class DebugLoggingError(Exception):
    """Custom debug logging exception"""

    def __init__(self, msg=None):
        self.message = msg if msg is not None else "Logging failure"
        super().__init__(self.message)


# ----------------------------------------------------------------------------
class ActionType(enum.Enum):
    """Enumeration of encoding actions"""

    NONE = 0  # Do nothing (i.e. use defaults)
    INLINE = 1  # Information encoded inline in URL
    FILE = 2  # Information is provided from provided file path
    SCRIPT = 3  # Information returned via stdout of provided executable


# ----------------------------------------------------------------------------
class DebugLevel(enum.IntEnum):
    """Required debug verbosity level to print debug message"""

    ALWAYS = 0           # Always print this message
    TEST = 1             # Test output reuired for a test
    DEBUGGING = 2        # Increased output for debugging purposes
    VERBOSE = 3          # Verbose debugging output
    CHATTY = 4           # Very chatty debugging output

# ----------------------------------------------------------------------------
PROGRAM_NAME = "debug_plugin.py"
PLUGIN_VERSION = "1.1.0"
SUPPORTED_SCHEMAS = "debug,decode,encoded"
# What Plugin protocol version does this speak
PROTOCOL_VERSION = 4

DEFAULT_HOSTNAME = "default.test.hostname"

OPT_ERR_PARAMETER = "parameter"
OPT_ERR_RESOLUTION = "resolution"
OPT_ERR_CONTACT = "contact"
OPT_ERR_AUTH = "authorization"
OPT_ERR_SPECIFICATION = "specification"
OPT_ERR_XFER = "transfer"
OPT_GENERAL = "general"
OPT_DELETE = "delete"

OPTION_DEFAULTS = {
    OPT_GENERAL: {},
    OPT_DELETE: [],
    OPT_ERR_PARAMETER: {
        "ErrorCode": -1,
        "ErrorString": "Invalid plugin parameters",
        "PluginVersion": PLUGIN_VERSION,
        "PluginLaunched": True,
    },
    OPT_ERR_RESOLUTION: {
        "ErrorCode": -2,
        "ErrorString": "Failed to resolve server name to contact",
        "FailedName": DEFAULT_HOSTNAME,
        "FailureType": "Definitive",
    },
    OPT_ERR_CONTACT: {
        "ErrorCode": -3,
        "ErrorString": "Failed to contact server",
        "FailedServer": DEFAULT_HOSTNAME,
    },
    OPT_ERR_AUTH: {
        "ErrorCode": -4,
        "ErrorString": "Failed authorization with server",
        "FailedServer": DEFAULT_HOSTNAME,
        "FailureType": "Authorization",
        "ShouldRefresh": True,
    },
    OPT_ERR_SPECIFICATION: {
        "ErrorCode": -5,
        "ErrorString": "Specified file DNE",
        "FailedServer": DEFAULT_HOSTNAME,
    },
    OPT_ERR_XFER: {
        "ErrorCode": -6,
        "ErrorString": "Failed to transfer file",
        "FailedServer": DEFAULT_HOSTNAME,
        "FailureType": "TooSlow",
    },
}


ACTION_TYPE_STR = {
    ActionType.NONE: "DetailsFromDefaults",
    ActionType.INLINE: "DetailsFromInline",
    ActionType.FILE: "DetailsFromFile",
    ActionType.SCRIPT: "DetailsFromScript",
}


DATETIME_TIMESTAMP_FORMAT = "%m/%d/%Y %H:%M:%S"
DEBUG_LEVEL = 0


# ----------------------------------------------------------------------------
def timestamp(msg: str) -> str:
    """Prepend message string with date timestamp"""
    timestamp = datetime.datetime.now().strftime(DATETIME_TIMESTAMP_FORMAT)
    ret = f"{timestamp} {msg}"
    return ret.strip()


# ----------------------------------------------------------------------------
def error(msg: str) -> None:
    """Write error message to standard error"""
    print(timestamp(msg), file=sys.stderr)


# ----------------------------------------------------------------------------
def debug(level: DebugLevel, msg: str) -> None:
    """Write debugging message to standard output depending on set verbosity level"""
    if DEBUG_LEVEL >= level:
        print(timestamp(msg), file=sys.stdout)


# ----------------------------------------------------------------------------

def set_debug_level(level: Union[int, str]) -> None:
    """Scoped function for changing the debug level on the fly"""
    global DEBUG_LEVEL

    if isinstance(level, int):
        # Just use this value
        pass
    elif isinstance(level, str):
        # Convert string to integer level
        # Either specifying log level by name (i.e. ALWAYS, TEST, CHATTY, etc.)
        # or by an integer value
        verbosity = None
        for log_level in DebugLevel:
            if log_level.name.upper() == level.upper():
                verbosity = log_level.value
                break

        if verbosity is None:
            try:
                verbosity = int(level)
            except:
                raise DebugLoggingError(f"Set Logging Level Failure: Invalid log level provided: {level}")

        level = verbosity
    else:
        invalid_type = type(level)
        raise DebugLoggingError(f"Set Logging Level Failure: Invalid level type ({invalid_type}). Expect int or str")

    if DEBUG_LEVEL != level:
        debug(DebugLevel.ALWAYS, f"Changing logging level from {DEBUG_LEVEL} to {level}")
        DEBUG_LEVEL = level


# ----------------------------------------------------------------------------
def format_error(error: Exception) -> str:
    """Format an exception into error string"""
    return "{0}: {1}".format(type(error).__name__, str(error))


# ----------------------------------------------------------------------------
def get_error_dict(error: Exception, url: str = "", parse_failure: bool = False) -> dict:
    """Return runtime plugin failure ClassAd"""
    error_string = format_error(error)
    error_dict = {
        "TransferSuccess": False,
        "TransferError": error_string,
        "TransferUrl": url,
        "IsParseFailure" : parse_failure,
    }

    return error_dict


# ----------------------------------------------------------------------------
# NOTE: This plug-in's URLs implicitly aren't allowed to have an empty host
#       portion and/or an absolute path after the `protocol://` part.
#       parse_portion() argument (url) is expected to be any and all characters
#       proceeding the first slash `protocol://command/url-portions`.
def parse_portion(url: str):
    """Parse out and yield the next portion of the URL to process"""
    option = ""
    details = ""
    action = ActionType.NONE
    in_brackets = 0
    trailing = False
    specifier = None

    # Parse URL one character at a time
    for c in url:
        # Always expect a '[' after an action specifier ('~','#')
        if specifier is not None and c != "[":
            raise DebugUrlParseError(f"Special file path delimeter token '{specifier}' not followed by '['")
        else:
            specifier = None

        # If not in brackets and '/' then yield stored information and rest
        if in_brackets == 0 and c == "/":
            yield (option, action, details)
            option = details = ""
            action = ActionType.NONE
            trailing = False

        # Ignore all information post closing bracket ']' This allows
        # `keyword[info]ignored/keyword` to produce multiple portions
        elif trailing:
            continue

        # Process start of specified encoding details
        elif c == "[":
            # If no action type has been provided then info is inline
            if action == ActionType.NONE:
                action = ActionType.INLINE
            # Count number of open brackets for nested information
            in_brackets += 1
            if action == ActionType.INLINE:
                details += c

        # Process end of specified encoding details
        elif c == "]":
            # Invalid URL if no start character ('[') found
            if in_brackets <= 0:
                raise DebugUrlParseError("Invalid URL has close bracket with no open.")
            # Found a pair of brackets '[]' reduce count
            in_brackets -= 1
            # If all bracket pairs found ignore trailing characters
            if in_brackets == 0:
                trailing = True
            if action == ActionType.INLINE:
                details += c

        # Store the details if in brackets (i.e. found open but not close)
        elif in_brackets > 0:
            details += c

        # Tilde specifies details is a file path to read
        elif c == "~":
            action = ActionType.FILE
            specifier = c

        # Hashtag specifies details is an executable to run
        elif c == "#":
            action = ActionType.SCRIPT
            specifier = c

        # If here then must be the start of the portion which is the sub option
        else:
            option += c

    # Went through full URL and has a mismatch of brackets
    if in_brackets != 0:
        raise DebugUrlParseError("Invalid URL has open bracket with no close.")

    # yield the final portion information
    yield (option, action, details)


# ----------------------------------------------------------------------------
def replace_white_space(url: str, seq: str) -> str:
    """Replace non-escaped charater sequence with a space"""
    debug(DebugLevel.CHATTY, f"Replacing '{seq}' with whitespaces in '{url}'")
    return url.replace(f"\\{seq}", "{**temp-replacement**}").replace(seq, " ").replace("{**temp-replacement**}", f"\\{seq}")

# ----------------------------------------------------------------------------
def extract_details(action: ActionType, details: str) -> dict:
    """Extract encoded information from a string"""
    ad = {}

    # Action type is none so do nothing (use default)
    if action == ActionType.NONE:
        debug(DebugLevel.CHATTY, "Using default transfer attempt ClassAd")
        pass

    # Action type is inline ClassAd
    elif action == ActionType.INLINE:
        try:
            # Replace any '+' w/ spaces (%20's) already replaced and attempt to parse ClassAd
            details = replace_white_space(details, "+")
            debug(DebugLevel.VERBOSE, f"Updating transfer attempt with inline ClassAd: '{details}'")
            inline_ad = classad.ClassAd(details)
            ad.update(inline_ad)
        except Exception as e:
            raise DebugUrlParseError(f"Failed to parse inline information '{details}' into ClassAd: {e}")

    # Action type is file
    elif action == ActionType.FILE:
        debug(DebugLevel.VERBOSE, f"Updating transfer attempt with contents from {details}")
        try:
            # Attempt to parse one ClassAd from the file
            with open(details, "r") as f:
                f_ad = classad.parseOne(f.read())
                ad.update(f_ad)
        except Exception as e:
            raise DebugUrlParseError(f"Failed to parse file '{details}' into ClassAd: {e}")

    # Action type is a script
    elif action == ActionType.SCRIPT:
        try:
            # Replace any '::' w/ spaces (%20's) already replaced
            details = replace_white_space(details, "::")
            debug(DebugLevel.VERBOSE, f"Updating transfer attempt with output from {details}")

            # Attempt to execute the script. Non-zero exit code is failure
            p = subprocess.run(details.split(" "), stdout=subprocess.PIPE)
            if p.returncode != EXIT_SUCCESS:
                raise RuntimeError(f"Failed to excute script")

            # Attempt to parse one ClassAd from stdout of script
            output = p.stdout.rstrip().decode()
            debug(DebugLevel.CHATTY, f"Script output:\n{output}")

            p_ad = classad.parseOne(output)
            ad.update(p_ad)
        except Exception as e:
            raise DebugUrlParseError(f"Failed to parse script '{details}' output into ClassAd: {e}")

    # Unknown action type...
    else:
        raise DebugUrlParseError(f"Unknown specified option action: {action}")

    return ad


# ----------------------------------------------------------------------------
class DebugPlugin:
    def __init__(self):
        self.exit_code = EXIT_SUCCESS
        self.sub_cmd_counts = {
            OPT_ERR_PARAMETER: 0,
            OPT_ERR_RESOLUTION: 0,
            OPT_ERR_CONTACT: 0,
            OPT_ERR_AUTH: 0,
            OPT_ERR_SPECIFICATION: 0,
            OPT_ERR_XFER: 0,
            OPT_GENERAL: 0,
            OPT_DELETE: 0,
        }
        self.details_counts = {
            ActionType.NONE: 0,
            ActionType.INLINE: 0,
            ActionType.FILE: 0,
            ActionType.SCRIPT: 0,
        }

    def reset_per(self) -> None:
        """Reset per URL statistics/data"""
        self.sub_cmd_counts = {key: 0 for key in self.sub_cmd_counts.keys()}
        self.details_counts = {key: 0 for key in self.details_counts}

    def get_dev_data(self) -> dict:
        """Return nested ClassAd of data pertaining to plugin execution"""
        data = {
            "CurrentPluginExitCode": self.exit_code,
        }
        data.update({f"Num{key.capitalize()}Option": val for key, val in self.sub_cmd_counts.items() if val > 0})
        data.update({ACTION_TYPE_STR[key]: val for key, val in self.details_counts.items() if val > 0})
        return data

    def process_url(self, url: str) -> dict:
        """Process full URL for encoded information"""
        info = {}
        attempts = []
        err_msg = "URL transfer encoded to fail"

        # Loop through custom parsed URL portions
        for option, action, details in parse_portion(url):
            # Replace all %20 w/ spaces
            details = details.replace("%20", " ")
            # Check if portion has specified inner option
            if option.lower() not in OPTION_DEFAULTS:
                debug(DebugLevel.VERBOSE, f"Skipping unknown encoded option {option}")
                continue

            # Copy default option ad
            ad = OPTION_DEFAULTS.get(option.lower()).copy()

            # Update internal statistics
            self.sub_cmd_counts[option.lower()] += 1
            self.details_counts[action] += 1

            # If 'general' option: update main result ad
            if re.match(OPT_GENERAL, option, re.I):
                debug(DebugLevel.CHATTY, "Updating general information for transfer result ClassAd")
                info.update(extract_details(action, details))
                continue

            # If 'delete' option: delete specified keys from main result ad
            if re.match(OPT_DELETE, option, re.I):
                debug(DebugLevel.CHATTY, "Deleting information in transfer result ClassAd")
                # Expects: delete[attr{,attr,...}]
                if action != ActionType.INLINE or details.strip("[").strip("]") == "":
                    raise DebugUrlParseError(f"Invalid {OPT_DELETE} option specified: No inline details.")
                for attr in details[1:-1].split(","):
                    if attr in info:
                        debug(DebugLevel.VERBOSE, f"Removing {attr} from transfer result ClassAd")
                        del info[attr]
                continue

            # Set the ErrorType attr in failure ad
            ad["ErrorType"] = option.lower().capitalize()
            # ad["DeveloperData"] = {}

            # Update this ad with encoded information
            ad.update(extract_details(action, details))

            # Add this ad to list of failure attempts
            attempts.append(ad)

        # If failure attempts have been specified then add ads to result ad
        if len(attempts) > 0:
            debug(DebugLevel.DEBUGGING, f"URL transfer encoded with {len(attempts)} failure attempts")
            info["TransferErrorData"] = attempts
            info["TransferError"] = err_msg  # We should do something better here

        return info

    def transfer_file(self, url: str, local_file: str, upload: bool) -> Tuple[dict, str]:
        """Mock transfer a file to produce return ClassAd from information encoded from URL"""
        self.reset_per()
        now = int(time.time())

        schema, data = url.split("://")

        # Default result ad
        RESULT = {
            "TransferProtocol": schema,
            "TransferType": "upload" if upload else "download",
            "TransferFileName": local_file,
            "TransferFileBytes": 0,
            "TransferTotalBytes": 0,
            "TransferStartTime": now,
            "TransferEndTime": now,
            "ConnectionTimeSeconds": 0,
            "TransferUrl": url,
        }

        # And a result string, only used when simulating an unparseable ad for EXITBADAD command
        RESULT_STR = ""

        try:
            # Parse main command from start of URL: cmd/.../...
            cmd, info = data.split("/", 1) if "/" in data else (data, "")
            cmd = cmd.upper()

            # Check main command

            # SLEEP (debug://sleep/<time>) -> Sleep for specified time (likely to invoke timeout)
            if cmd == "SLEEP":
                try:
                    timeout = int(info.split("/")[0])
                except Exception as e:
                    raise DebugUrlParseError(f"Invalid {cmd} command: {e}")
                debug(DebugLevel.DEBUGGING, f"Sleeping for {timeout} seconds...")
                time.sleep(timeout)
                debug(DebugLevel.CHATTY, "Finished sleeping!")
                RESULT["TransferSuccess"] = True

            # SIGNAL (debug://signal/SIGTERM) -> Send a signal to self
            elif cmd == "SIGNAL":
                value = info.split("/")[0]
                SIGNAL = getattr(signal, value, signal.SIGTERM)
                debug(DebugLevel.DEBUGGING, f"Sending signal {SIGNAL.name} to self...")
                os.kill(os.getpid(), SIGNAL)
                error(f"Failed to send signal {SIGNAL.name} to self!!!")

            # ERROR (debug://error/...) -> Specify a transfer failure with encoded return ad information
            elif cmd == "ERROR":
                debug(DebugLevel.VERBOSE, f"Specified failed transfer encoded via '{info}'")
                RESULT["TransferSuccess"] = False
                RESULT.update(self.process_url(info))
                if self.exit_code == EXIT_SUCCESS:
                    self.exit_code = 1

            # SUCCESS (debug://success/...) -> Specify a successful transfer with encoded return ad information
            elif cmd == "SUCCESS":
                debug(DebugLevel.CHATTY, "Specified successfull transfer")
                RESULT["TransferSuccess"] = True
                RESULT.update(self.process_url(info))

            # EXIT (debug://exit/<code>) -> Inform plugin to exit right now with the specific code
            elif cmd == "EXIT" or cmd == "EXITNOAD" or cmd == "EXITBADAD":
                try:
                    # Limit exit code range 0-123 (negatives are turned into absolute values)
                    code = sorted([0, abs(int(info.split("/")[0])), 123])[1]
                except Exception as e:
                    raise DebugUrlParseError(f"Invalid {cmd} command: {e}")

                debug(DebugLevel.CHATTY, f"Specified exit code: {code}")
                # Only override exit code if no invalid URL's were parsed
                if self.exit_code != EXIT_INVALID_URL:
                    debug(DebugLevel.DEBUGGING, f"Changing exit code from {self.exit_code} to {code}")
                    self.exit_code = code

                RESULT["TransferSuccess"] = (self.exit_code == EXIT_SUCCESS)

            # ECHO (debug://echo/stdout/message)
            elif cmd == "ECHO":
                try:
                    stream, message = info.split("/", 1)
                    stream = stream.upper()

                    if stream not in ["STDOUT", "STDERR", "OUTPUT", "ERROR", "OUT", "ERR"]:
                        raise RuntimeError(f"Invalid echo stream provided: {stream}")

                    message = replace_white_space(message, "%20")
                    if stream in ["STDOUT", "OUT", "OUTPUT"]:
                        debug(DebugLevel.ALWAYS, message)
                    else:
                        error(message)
                except Exception as e:
                    raise DebugUrlParseError(f"Invalid {cmd} command: {e}")

            # VERBOSITY (debug://verbosity/3)
            elif cmd == "VERBOSITY":
                try:
                    set_debug_level(info.split("/")[0])
                except Exception as e:
                    raise DebugUrlParseError(f"Invalid {cmd} command: {e}")

            else:
                raise DebugUrlParseError(f"Unknown debug URL command '{cmd}'")

        # Catch all exceptions and fail due to an invalid encoded URL provided
        except Exception as e:
            error(f"Error: Failed to execute command ({cmd}): {e}")
            RESULT.update(get_error_dict(e, url, True))
            self.exit_code = EXIT_INVALID_URL

        # Add general debugging Developer Data to output
        RESULT["DeveloperData"] = self.get_dev_data()

        # Deal with EXITBADAD command which produces an unparseable ClassAd by writing a string instead of a dict (i.e. an ad) to the output file
        if cmd == "EXITBADAD":
            RESULT_STR = "This is not a ClassAd!!!" 

        # Deal with EXITNOAD command which produces no output at all by returning an empty dict so that no file is created
        if cmd == "EXITNOAD":
            RESULT = {}

        return RESULT, RESULT_STR


# ----------------------------------------------------------------------------
def recursive(l: list) -> None:
    """Recursively loop through nested lists"""
    for item in l:
        if type(item) is list:
            yield from recursive(item)
        else:
            yield item


# ----------------------------------------------------------------------------
def parse_args() -> None:
    """Function to handle all CLI argument parsing"""

    parser = argparse.ArgumentParser(
        prog=PROGRAM_NAME,
        description=textwrap.dedent(
            f"""
            Center for High Throughput Computing, Computer Sciences
            Department, University of Wisconsin-Madison, Madison, WI, US

            Debug Plugin
                This is a custom HTCondor file transfer plugin that parses
                provided URL's to determine how the plugin writes output
                data and exits to provide a testing infrastucture for how
                the file transfer acts provided specific scenarios.
            """
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "-classad",
        dest="classad",
        action="store_true",
        default=False,
        help="Print a ClassAd containing the capablities of this file transfer plugin.",
    )

    parser.add_argument(
        "-infile",
        dest="infile",
        action="store",
        type=str,
        default=None,
        help="Input ClassAd file",
    )

    parser.add_argument(
        "-outfile",
        dest="outfile",
        action="store",
        type=str,
        default=None,
        help="Output ClassAd file",
    )

    parser.add_argument(
        "-upload",
        dest="upload",
        action="store_true",
        default=False,
        help="Indicates this transfer is an upload (default is download).",
    )

    parser.add_argument(
        "-test",
        dest="testing",
        action="append",
        nargs="+",
        default=None,
        help="Return plugin result dictionary for each provided test url.",
    )

    parser.add_argument(
        "-v",
        "-verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="Increase plugin verbosity",
    )

    args = parser.parse_args()

    # If verbosity flag specified then use this level over the environment set level (if one)
    if args.verbosity > 0:
        set_debug_level(args.verbosity)

    # Special -test flag directly parses provided url and prints resulting ClassAd
    if args.testing is not None:
        plugin = DebugPlugin()
        for url in recursive(args.testing):
            debug(DebugLevel.DEBUGGING, f"{url} yields:")
            result, result_string = plugin.transfer_file(url, "TESTING-DUMMY", args.upload)
            try:
                if result_string:
                    print(result_string)
                else:   
                    print(classad.ClassAd(result))
            except Exception as e:
                error(f"Error: Failed to convert results to ClassAd: {e}")
                error(result)
        sys.exit(plugin.exit_code)

    # Ensure either -classad or (-infile and -outfile) specified
    if not args.classad and None in [args.infile, args.outfile]:
        parser.print_usage()
        error(f"{PROGRAM_NAME}: error: invalid args: expects either -classad or both -infile and -outfile")
        raise RuntimeError

    # Return capabilities ClassAd and exit
    if args.classad:
        capabilities = {
            "MultipleFileSupport": True,
            "PluginType": "FileTransfer",
            "SupportedMethods": SUPPORTED_SCHEMAS,
            "Version": PLUGIN_VERSION,
            "ProtocolVersion": PROTOCOL_VERSION,
        }
        sys.stdout.write(classad.ClassAd(capabilities).printOld())
        sys.exit(EXIT_SUCCESS)

    return args


# ----------------------------------------------------------------------------
def main() -> None:
    try:
        args = parse_args()
    except Exception as err:
        error(f"Error: Failed to parse arguments: {err}")
        sys.exit(EXIT_SETUP_FAILURE)

    plugin = DebugPlugin()

    # Parse in the classads stored in the input file.
    # Each ad represents a single file to be transferred.
    try:
        infile_ads = classad.parseAds(open(args.infile, "r"))
    except Exception as err:
        error(f"Error: Failed to parse infile {args.infile} ClassAds: {err}")
        try:
            with open(args.outfile, "w") as outfile:
                outfile_dict = get_error_dict(err)
                outfile.write(str(classad.ClassAd(outfile_dict)))
        except Exception as deep_err:
            error(f"Error: Failed to write failure output ClassAd: {deep_err}")
        sys.exit(EXIT_SETUP_FAILURE)

    # Now iterate over the list of classads and perform the transfers.
    try:
        with open(args.outfile, "w") as outfile:
            for ad in infile_ads:

                # Non-transfer ClassAd
                if "URL" not in ad or "LocalFileName" not in ad:
                    debug(DebugLevel.TEST, f"Non-transfer ClassAd discovered:\n{ad}")
                    continue

                # Dump v4 protocol Plugin Data
                if 'PluginData' in ad:
                    data = ad.get("PluginData")
                    debug(DebugLevel.TEST, f"Plugin data:\n{data}")

                for key in ad.keys():
                    if key.endswith('_PluginData'):
                        data = ad.get(key)
                        debug(DebugLevel.TEST, f"Specific Plugin Data (key = {key}):\n{data}")

                # Process transfer ClassAd for action
                try:
                    url = ad["Url"]
                    outfile_dict, outfile_string = plugin.transfer_file(url, ad["LocalFileName"], args.upload)
                    if (len(outfile_string) > 0):
                        outfile.write(outfile_string)
                    elif len(outfile_dict) > 0:   # Note: do not create a file if dict empty since some commands (i.e. EXITNOAD) are meant to simulate failures that don't produce a result ad
                        outfile.write(str(classad.ClassAd(outfile_dict)))
                except Exception as err:
                    error(f"Error: Failure to process {url}: {err}")
                    try:
                        outfile_dict = get_error_dict(err, url=url)
                        outfile.write(str(classad.ClassAd(outfile_dict)))
                    except Exception as deep_err:
                        error(f"Error: Failed to write failure output ClassAd: {deep_err}")
                    sys.exit(EXIT_SETUP_FAILURE)

    except Exception as e:
        error(f"Error: Setup failure: {e}")
        sys.exit(EXIT_SETUP_FAILURE)

    sys.exit(plugin.exit_code)


# ----------------------------------------------------------------------------
def process_env() -> None:
    """Process environment variables"""
    global PROTOCOL_VERSION
    global DATETIME_TIMESTAMP_FORMAT

    env = os.environ

    try:
        PROTOCOL_VERSION = int(env.get('DEBUG_PLUGIN_PROTOCOL_VERSION', PROTOCOL_VERSION))
    except Exception as e:
        error(f"WARNING: Failed to process environment variable DEBUG_PLUGIN_PROTOCOL_VERSION : {e}")

    try:
        tmp = env.get("DEBUG_PLUGIN_TIMESTAMP_FORMAT", DATETIME_TIMESTAMP_FORMAT)
        if bool(datetime.datetime.now().strftime(tmp)):
            DATETIME_TIMESTAMP_FORMAT = tmp
        else:
            error(f"WARNING: Enviroment variable DEBUG_PLUGIN_TIMESTAMP_FORMAT has invalid value '{tmp}'")
    except Exception as e:
        error(f"WARNING: Failed to process evironment variable DEBUG_PLUGIN_TIMESTAMP_FORMAT: {e}")

    try:
        set_debug_level(env.get("DEBUG_PLUGIN_VERBOSITY", "ALWAYS"))
    except Exception as e:
        error(f"WARNING: Failed to process environment variable DEBUG_PLUGIN_VERBOSITY: {e}")

# ----------------------------------------------------------------------------
if __name__ == "__main__":
    process_env()
    main()



#!/usr/bin/env python3

import argparse
import classad2 as classad
import gwdatafind
import io
import json
import os
import pathlib
import posixpath
import pycurl
import shutil
import socket
import sys
import time

try:
    from urllib.parse import urlparse # Python 3
except ImportError:
    from urlparse import urlparse # Python 2

DEFAULT_TIMEOUT = 30
GWDATA_PLUGIN_VERSION = '1.0.0'

def print_help(stream = sys.stderr):
    help_msg = '''Usage: {0} -infile <input-filename> -outfile <output-filename>
       {0} -classad

Options:
  -classad                    Print a ClassAd containing the capablities of this
                              file transfer plugin.
  -infile <input-filename>    Input ClassAd file
  -outfile <output-filename>  Output ClassAd file
  -upload                     Not supported
'''
    stream.write(help_msg.format(sys.argv[0]))

def print_capabilities():
    capabilities = {
         'MultipleFileSupport': True,
         'PluginType': 'FileTransfer',
         'SupportedMethods': 'gwdata',
         'Version': GWDATA_PLUGIN_VERSION,
    }
    sys.stdout.write(classad.ClassAd(capabilities).printOld())

def parse_args():
    '''The optparse library can't handle the types of arguments that the file
    transfer plugin sends, the argparse library can't be expected to be
    found on machines running EL 6 (Python 2.6), and a plugin should not
    reach outside the standard library, so the plugin must roll its own argument
    parser. The expected input is very rigid, so this isn't too awful.'''

    # The only argument lists that are acceptable are
    # <this> -classad
    # <this> -infile <input-filename> -outfile <output-filename>
    # <this> -outfile <output-filename> -infile <input-filename>
    if not len(sys.argv) in [2, 5, 6]:
        print_help()
        sys.exit(1)

    # If -classad, print the capabilities of the plugin and exit early
    if (len(sys.argv) == 2) and (sys.argv[1] == '-classad'):
        print_capabilities()
        sys.exit(0)

    # Upload is not supported
    if '-upload' in sys.argv[1:]:
        print("Upload is not supported. You should not be seeing this.")
        sys.exit(1)

    # -infile and -outfile must be in the first and third position
    if not (
            ('-infile' in sys.argv[1:]) and
            ('-outfile' in sys.argv[1:]) and
            (sys.argv[1] in ['-infile', '-outfile']) and
            (sys.argv[3] in ['-infile', '-outfile']) and
            (len(sys.argv) == 5)):
        print_help()
        sys.exit(1)
    infile = None
    outfile = None
    try:
        for i, arg in enumerate(sys.argv):
            if i == 0:
                continue
            elif arg == '-infile':
                infile = sys.argv[i+1]
            elif arg == '-outfile':
                outfile = sys.argv[i+1]
    except IndexError:
        print_help()
        sys.exit(1)

    return {'infile': infile, 'outfile': outfile}

def format_error(error):
    return '{0}: {1}'.format(type(error).__name__, str(error))

def get_error_dict(error, url = ''):
    error_string = format_error(error)
    error_dict = {
        'TransferSuccess': False,
        'TransferError': error_string,
        'TransferUrl': url,
    }

    return error_dict

class GWDataPlugin:

    #def __init__(self):

    def get_urls(self, host, args):

        urls = []

        # Parse the input arguments
        for arg in args:
            attr, value = arg.split("=")
            if attr == "observatory": observatory = value
            if attr == "type": type = value
            if attr == "s": start_frame = value
            if attr == "e": end_frame = value

        # Retrieve the list of URLs
        try:
            urls = gwdatafind.find_urls(
                host=host,
                site=observatory,
                frametype=type,
                gpsstart=int(start_frame),
                gpsend=int(end_frame)
            )
        except Exception as e:
            print("Error retrieving gwdatafind URLs: " + str(sys.exc_info()[0]) + ": " + str(e))

        return urls

    def create_cache(self, args, urls):

        metadata_filename = "metadata.txt"

        # Parse the input arguments
        for arg in args:
            attr, value = arg.split("=")
            if attr == "e": end_frame = int(value)
            if attr == "cache": cache = value
            if attr == "metadata_file": metadata_filename = value

        metadata_file = open(metadata_filename, "w")

        if cache == "lal" or cache == "lal-cache":
            for url in urls:
                file_name = url.split("/")[-1]
                metadata_tokens = file_name.split("-")
                metadata_file.write("{0} {1} {2} {3} {4}\n".format(
                    metadata_tokens[0],
                    metadata_tokens[1],
                    metadata_tokens[2],
                    metadata_tokens[3].split(".")[0],
                    pathlib.Path().absolute()
                ))
        elif cache == "frame" or cache == "frame-cache":
            for url in urls:

                # Filename and metadata specific to this url
                file_name = url.split("/")[-1]
                metadata_tokens = file_name.split("-")

                # If this is the first url in the list, set a couple gps tracking variables
                if url == urls[0]:
                    gps_start = int(metadata_tokens[2])
                    gps_counter = gps_start
                else:
                    gps_current = int(metadata_tokens[2])
                    gps_interval = int(metadata_tokens[3].split(".")[0])
                    gps_counter += gps_interval

                    # If this url represents a gap, write data to file
                    if (gps_current != gps_counter) or (url == urls[-1]):
                        gps_end = gps_counter
                        if url == urls[-1]:
                            gps_end = end_frame

                        metadata_file.write("{0} {1} {2} {3} {4} {5}\n".format(
                            metadata_tokens[0],
                            metadata_tokens[1],
                            gps_start,
                            gps_end,
                            gps_interval,
                            pathlib.Path().absolute()
                        ))
                        # Now reset reset gps tracking variables
                        gps_start = gps_current
                        gps_counter = gps_start
        else:
            metadata_file.write("Error: {0} is not a valid cache type. Allowed types are \"lal\", \"lal-cache\", \"frame\" and \"frame-cache\".\n".format(cache))

        metadata_file.close()


    def download_data(self, url):

        transfer_stats = []
        
        gwdatafind_server = url[(url.find("://") + 3):url.find("?")]
        gwdata_args = url[(url.find("?") + 1):].split("&")

        # First retrieve a list of urls to download from the gwdatafind server
        gwdata_urls = self.get_urls(gwdatafind_server, gwdata_args)

        # Now transfer each url
        curl = pycurl.Curl()
        for gwdata_url in gwdata_urls:

            start_time = time.time()

            # Setup the transfer
            transfer_success = True
            transfer_error = ""
            file_name = gwdata_url.split("/")[-1]
            file_size = 0
            file = open(file_name, "wb")

            # Use pycurl to actually grab the file
            curl.setopt(curl.URL, gwdata_url)
            curl.setopt(curl.WRITEDATA, file)
            try:
                curl.perform()
                file_size = file.tell()
                file.close()
            except Exception as error:
                transfer_success = False
                errno, errstr = error.args
                transfer_error = str(errstr + " (Error " + str(errno) + ")")

            end_time = time.time()

            # Add statistics for this transfer
            this_transfer_stats = {
                'TransferSuccess': transfer_success,
                'TransferProtocol': 'gwdata',
                'TransferType': 'download',
                'TransferFileName': file_name,
                'TransferFileBytes': file_size,
                'TransferTotalBytes': file_size,
                'TransferStartTime': int(start_time),
                'TransferEndTime': int(end_time),
                'ConnectionTimeSeconds': end_time - start_time,
                'TransferUrl': gwdata_url,
            }
            if transfer_error:
                this_transfer_stats['TransferError'] = transfer_error

            transfer_stats.append(this_transfer_stats)

            # If this transfer failed, exit immediately
            if not transfer_success:
                curl.close()
                return transfer_stats, False

        # Check if the args list is requesting a cache file; if so, create it
        if "cache" in url:
            self.create_cache(gwdata_args, gwdata_urls)


        # Looks like everything downloaded correctly. Exit success.
        curl.close()
        return transfer_stats, True



if __name__ == '__main__':

    # Start by parsing input arguments
    try:
        args = parse_args()
    except Exception:
        sys.exit(1)

    gwdata = GWDataPlugin()

    # Parse in the classads stored in the input file. 
    # Each ad represents a single file to be transferred.
    try:
        infile_ads = classad.parseAds(open(args['infile'], 'r'))
    except Exception as err:
        try:
            with open(args['outfile'], 'w') as outfile:
                outfile_dict = get_error_dict(err)
                outfile.write(str(classad.ClassAd(outfile_dict)))
        except Exception:
            pass
        sys.exit(1)
    
    # Now iterate over the list of classads and perform the transfers.
    try:
        with open(args['outfile'], 'w') as outfile:
            for ad in infile_ads:
                try:
                    result_ads, transfer_success = gwdata.download_data(ad['Url'])
                    # Output the results to file
                    for ad in result_ads:
                        outfile.write(str(classad.ClassAd(ad)))

                    if not transfer_success:
                        sys.exit(1)

                except Exception as err:
                    print(err)
                    try:
                        outfile_dict = get_error_dict(err, url = ad['Url'])
                        outfile.write(str(classad.ClassAd(outfile_dict)))
                    except Exception:
                        pass
                    sys.exit(1)

    except Exception:
        sys.exit(1)



#!/usr/bin/env pytest

import logging
import htcondor2 as htcondor

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class TestHTCondorSubmitConstructor:
    def test_passing_none(self):
        # This used to segfault in version 1.
        try:
            s = htcondor.Submit({"attribute": None})
        except TypeError:
            pass



from typing import Union

from ._log_level import LogLevel
from .htcondor2_impl import _py_dprintf

def _log(level : Union[LogLevel, int], message : str):
    """
    Log a message using the HTCondor logging subsystem.

    :param level:  Specify multiple :class:`LogLevel` values using bitwise-or.
    :param message:  The message to log.
    """
    if not (isinstance(level, LogLevel) or isinstance(level, int)):
        raise TypeError("level must a LogLevel or a bitwise-or of such")
    _py_dprintf(int(level), message + "\n")




import os
import glob
import json
import tempfile

from typing import List

import scitokens
import htcondor2 as htcondor
import classad2 as classad

from credmon.CredentialMonitors.OAuthCredmon import OAuthCredmon
from credmon.utils import atomic_rename


_KEY_ALGO_PREAMBLE = {
    "RS256": "BEGIN RSA PRIVATE KEY",
    "ES256": "BEGIN EC PRIVATE KEY",
}


class TokenInfo:
    def __init__(
            self,
            sub: str,
            scopes: List[str],
            audience: List[str],
            profile: str,
            ):
        self.sub = sub
        self.scopes = scopes
        self.audience = audience
        self.profile = profile


class LocalCredmon(OAuthCredmon):
    """
    This credential monitor class provides the ability to self-sign SciTokens
    without needing to access a remote service; only useful when the credd host has
    a copy of the private signing key.
    """

    use_token_metadata = False

    @property
    def credmon_name(self):
        return "LOCAL"

    def __init__(self, provider, *args, **kwargs):
        super(LocalCredmon, self).__init__(*args, **kwargs)
        self.provider = provider
        self.token_issuer = None
        self.authz_template = r"read:/user/{username} write:/user/{username}"
        self.authz_template_expr = None
        self.authz_group_template = None
        self.authz_group_mapfile = None
        self.authz_group_requirement = None
        self.token_lifetime = 60*20
        self.token_use_json = True
        self.token_aud = ""
        self.token_ver = "scitoken:2.0"
        if htcondor != None:
            self._private_key_location = self.get_credmon_config("PRIVATE_KEY", "/etc/condor/scitokens-private.pem")
            if self._private_key_location != None and os.path.exists(self._private_key_location):
                with open(self._private_key_location, 'r') as private_key:
                    self._private_key = private_key.read()
                self.private_key_id = self.get_credmon_config("KEY_ID", "local")
                self.private_key_algo = self.get_credmon_config("PRIVATE_KEY_ALGORITHM", "ES256")
            else:
                self.log.error(f"{self.credmon_name}_CREDMON_PRIVATE_KEY specified at {self._private_key_location}, but key not found or not readable")
            self.token_issuer = self.get_credmon_config("ISSUER", self.token_issuer)
            self.authz_template = self.get_credmon_config("AUTHZ_TEMPLATE", self.authz_template)
            self.authz_template_expr = self.get_credmon_config("AUTHZ_TEMPLATE_EXPR", self.authz_template_expr)
            self.authz_group_mapfile = self.get_credmon_config("AUTHZ_GROUP_MAPFILE", self.authz_group_mapfile)
            self.authz_group_template = self.get_credmon_config("AUTHZ_GROUP_TEMPLATE", self.authz_group_template)
            self.authz_group_requirement  = self.get_credmon_config("AUTHZ_GROUP_REQUIREMENT", self.authz_group_requirement)
            self.token_lifetime = self.get_credmon_config("TOKEN_LIFETIME", self.token_lifetime)
            self.token_use_json = self.get_credmon_config("TOKEN_USE_JSON", self.token_use_json)
            self.token_aud = self.get_credmon_config("TOKEN_AUDIENCE", self.token_aud)
            self.token_ver = self.get_credmon_config("TOKEN_VERSION", self.token_ver)
        else:
            self._private_key_location = None
        if not self.token_issuer and htcondor:
            self.token_issuer = 'https://{}'.format(htcondor.param["FULL_HOSTNAME"])
        # Check that the key matches the configured algorithm
        if (self._private_key_location is not None) and (_KEY_ALGO_PREAMBLE.get(self.private_key_algo, "") not in self._private_key.split("\n")[0]):
            self.log.warning(f"{self.credmon_name}_CREDMON_PRIVATE_KEY must use {self.private_key_algo} algorithm")
            self.log.warning(f"{_KEY_ALGO_PREAMBLE.get(self.private_key_algo, '')} not found in key file")
            self.log.warning("Errors are likely to occur when attempting to serialize SciTokens")
        elif self.private_key_algo not in _KEY_ALGO_PREAMBLE:
            self.log.warning(f"New or unknown private key algorithm '{self.private_key_algo}'")
            self.log.warning("Errors may occur when attempting to serialize SciTokens")


    def get_credmon_config(self, config: str, default: str="") -> str:
        """
        Given a local credmon config variable `FOO`, provider `BAR`, and default `BAZ`,
        returns the first matching of:
        - Value of `LOCAL_CREDMON_BAR_FOO` in the condor config
        - Value of `LOCAL_CREDMON_FOO` in the condor config
        - Default value

        (where `LOCAL` is substituted with the current class's `credmon_name` attribute)
        """
        return htcondor.param.get(f"{self.credmon_name}_CREDMON_{self.provider}_{config}", htcondor.param.get(f"{self.credmon_name}_CREDMON_{config}", default))


    def generate_access_token_info(self, username: str, token_name: str) -> TokenInfo:
        """
        Determines what information should be in the token for a given username / provider
        """
        # Create scopes from user and group templates
        scopes = [self.authz_template.format(username=username)]
        if self.authz_template_expr:
            try:
                eval_context = classad.ClassAd({
                    "Username": username,
                    "Scopes": classad.ExprTree(self.authz_template_expr),
                })
                eval_scopes = eval_context.eval("Scopes")
                if not isinstance(eval_scopes, str):
                    raise TypeError(f"ClassAd expression '{self.authz_template_expr}' did not evaluate to a string")
                scopes.append(eval_scopes)
            except Exception:
                self.log.exception(f"Not adding scopes from {self.credmon_name}_CREDMON_AUTHZ_TEMPLATE_EXPR due to exception")
        if self.authz_group_template:
            if self.authz_group_mapfile is None:
                self.log.warning(f"{self.credmon_name}_CREDMON_AUTHZ_GROUP_MAPFILE is undefined, cannot add group authorizations")
            else:
                try:
                    groups = get_user_groups(username, self.authz_group_mapfile)
                    self.log.info("Found {n} groups for user {username}".format(n=len(groups), username=username))
                    for groupname in groups:
                        scopes.append(self.authz_group_template.format(groupname=groupname))
                except IOError:
                    self.log.exception("Could not open {mapfile}, cannot add group authorizations".format(mapfile=self.authz_group_mapfile))

        profile = ""
        if self.token_ver:
            profile = self.token_ver

        # Convert the space separated list of audiences to a proper list
        # No aud is valid for scitokens:1.0 tokens.  Also, no resonable default.
        aud_list = [aud for aud in self.token_aud.strip().split() if aud]

        return TokenInfo(sub=username, scopes=scopes, profile=profile, audience=aud_list)

    def write_access_token(self, username: str, token_name: str, token_lifetime: str, serialized_token: str) -> bool:
        """
        Write a serialized access token to the credential directory.
        """
        if isinstance(serialized_token, bytes):
            serialized_token = serialized_token.decode()
        (tmp_fd, tmp_access_token_path) = tempfile.mkstemp(dir = self.cred_dir)
        with os.fdopen(tmp_fd, 'w') as f:
            if self.token_use_json:
                # use JSON if configured to do so, i.e. when
                # LOCAL_CREDMON_TOKEN_USE_JSON = True (default)
                f.write(json.dumps({
                    "access_token": serialized_token,
                    "expires_in":   int(token_lifetime),
                }))
            else:
                # otherwise write a bare token string when
                # LOCAL_CREDMON_TOKEN_USE_JSON = False
                f.write(serialized_token+'\n')

        access_token_path = os.path.join(self.cred_dir, username, token_name + '.use')

        # atomically move new file into place
        try:
            atomic_rename(tmp_access_token_path, access_token_path)
        except OSError as e:
            self.log.exception("Failure when writing out new access token to {}: {}.".format(
                access_token_path, str(e)))
            return False
        else:
            return True


    def refresh_access_token(self, username, token_name):
        """
        Create a SciToken at the specified path.
        """

        # Refuse to provide a token to users outside a specific group (if configured)
        if self.authz_group_requirement:
            if not self.authz_group_mapfile:
                self.log.error(f"Local credmon {self.provider} is configured to filter on group membership {self.authz_group_requirement} but {self.credmon_name}_CREDMON_{self.provider}_AUTHZ_GROUP_MAPFILE is undefined")
                return False
            try:
                groups = get_user_groups(username, self.authz_group_mapfile)
            except IOError:
                self.log.exception(f"Could not open {self.authz_group_mapfile}, cannot filter based on group membership")
                return False
            if self.authz_group_requirement not in groups:
                self.log.error(f"User {username} request for a token from provider {self.provider} is denied as {username} is not a member of group {self.authz_group_requirement}")
                self.log.error(f"Writing an empty {self.provider} token for {username}")
                return self.write_access_token(username, token_name, self.token_lifetime, "")

        info = self.generate_access_token_info(username, token_name)

        token = scitokens.SciToken(algorithm=self.private_key_algo, key=self._private_key, key_id=self.private_key_id)
        token.update_claims({'sub': info.sub})
        token.update_claims({'scope': " ".join(info.scopes)})

        # Only set the version if we have one.  No version is valid, and implies scitokens:1.0
        if info.profile == "wlcg:1.0" or info.profile == "wlcg":
            token.update_claims({"wlcg.ver": "1.0"})
        elif info.profile:
            token.update_claims({'ver': info.profile})

        if info.audience:
            token.update_claims({'aud': info.audience})
        elif self.token_ver.lower() == "scitoken:2.0":
            self.log.error(f'No "aud" claim, {self.credmon_name}_CREDMON_TOKEN_AUDIENCE must be set when requesting a scitoken:2.0 token')
            return False

        # Serialize the token and write it to a file
        try:
            serialized_token = token.serialize(issuer=self.token_issuer, lifetime=int(self.token_lifetime))
        except TypeError:
            self.log.exception("Failure when attempting to serialize a SciToken, likely due to algorithm mismatch")
            return False
        return self.write_access_token(username, token_name, self.token_lifetime, serialized_token)


    def process_cred_file(self, cred_fname):
        """
        Split out the file path to get username and base.
        Pass that data to the SciToken acquiring function.

        Format of cred_path should be:
        <cred_dir> / <username> / <provider>.top
        """
        # Take the cred_dir out of the cred_path
        base, _ = os.path.split(cred_fname)
        username = os.path.basename(base)

        if self.should_renew(username, self.provider):
            self.log.info('Found %s, acquiring SciToken and .use file', cred_fname)
            try:
                success = self.refresh_access_token(username, self.provider)
            except Exception as exc:
                success = False
                self.log.exception(f"Exception occurred when refreshing access token: {exc}")
            if success:
                self.log.info("Successfully renewed SciToken for user: %s", username)
            else:
                self.log.error("Failed to renew SciToken for user: %s", username)

    def scan_tokens(self):
        """
        Scan the credential directory for new credential requests.

        The LocalCredmon will look for files of the form `<username>/<provider>.top`
        and create the corresponding access token files, then invoke the parent OAuthCredmon
        method.
        """

        provider_glob = glob.glob(os.path.join(self.cred_dir, "*", f"{self.provider}.top"))
        self.log.debug(f"Found {len(provider_glob)} {self.provider} tokens to process")

        for file_name in provider_glob:
            self.log.debug(f"Processing {file_name}")
            self.process_cred_file(file_name)

        super(LocalCredmon, self).scan_tokens


def get_user_groups(username, mapfile):
    """Scan mapfile to find groups for username,
    returns a list of groups.

    Note that every call will read the current file,
    regardless of if HTCondor or the credmon have
    been reconfigured.
    
    Mapfiles have the format:
    * <userA> <group1>,<group2>,...,<groupN>
    * <userB> <group1>,<group2>,...,<groupN>
    """
    groups = []
    with open(mapfile) as f:
        for line in f:
            tokens = line.split()
            if len(tokens) < 3:
                continue
            if tokens[0] != "*":
                continue
            if tokens[1] == username:
                groups = tokens[2].split(",")
                break
    return groups



#!/usr/bin/env pytest

#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
    # simulate detection of type 4 (A100), 3 units
    GPU_SIM = 4,3
    GPU_DISCOVERY_EXTRA = $(GPU_DISCOVERY_EXTRA) -sim:$(GPU_SIM)
    use Feature : PartitionableSlot(1)
    use Feature : PartitionableSlot(2)
    SLOT_TYPE_2_BACKFILL = true
    STARTD_EVENTLOG = $(LOG)/EpEvents
    STARTD_DEBUG = $(STARTD_DEBUG) D_COMMAND:1
"""
#endtestreq

#import pytest
import logging
import htcondor2 as htcondor
import classad2 as classad

#from pathlib import Path
from ornithology import *


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def get_who_ads(default_condor):
    who_snap = default_condor.run_command(["condor_who", "-startd", "-snap", "-long"])

    # extract the condor_who ads and put them into a dict of dicts by type and name
    # where slot related ads will be keyed by the slot prefix (i.e. slot1_1)
    the_ads = {'StartD':{}, 'Machine':{}, 'Slot.Config':{}, 'Slot.State':{}, 'Machine.Extra':{}}
    for ad in classad.parseAds(who_snap.stdout):
        mytype = ad.get('MyType')
        name = ad.get('Name')
        if mytype != 'Machine.Extra' : name = name.split("@",maxsplit=1)[0]
        the_ads[mytype][name]=ad

    return the_ads


@action
def the_before_ads(default_condor, test_dir):
    return get_who_ads(default_condor)

@action
def the_after_ads(default_condor, test_dir):
    return get_who_ads(default_condor)

@action
def reconfig_and_wait(default_condor, test_dir):
    success = False
    default_condor.run_command(['condor_reconfig'])
    with default_condor.use_config():
        ep_eventlog = htcondor.param.get('STARTD_EVENTLOG')
        with htcondor.JobEventLog(ep_eventlog) as evl:
            for ev in evl.events(60):
                if ev.type == htcondor.JobEventType.EP_RECONFIG:
                    success = True
                    break

    return success


class TestBackfillGpus:

    def test_slots_have_gpus(self, the_before_ads):
        who_ads = the_before_ads
        slot1 = who_ads.get('Machine').get('slot1')
        slot2 = who_ads.get('Machine').get('slot2')
        expect_assign1 = 'GPU-a0223334,GPU-b1223334,GPU-c2223334'
        expect_assign2 = 'GPU-c2223334,GPU-b1223334,GPU-a0223334'
        assert (slot1.get('GPUs') == 3
            and slot1.get('TotalSlotGPUs') == 3
            and slot1.get('AssignedGPUs') == expect_assign1
            and slot2.get('GPUs') == 3
            and slot2.get('TotalSlotGPUs') == 3
            and slot2.get('AssignedGPUs') == expect_assign2)

    def test_reconfig_and_wait(self, reconfig_and_wait):
        assert reconfig_and_wait

    def test_slots_still_have_gpus(self, the_after_ads):
        who_ads = the_after_ads
        slot1 = who_ads.get('Machine').get('slot1')
        slot2 = who_ads.get('Machine').get('slot2')
        expect_assign1 = 'GPU-a0223334,GPU-b1223334,GPU-c2223334'
        expect_assign2 = 'GPU-c2223334,GPU-b1223334,GPU-a0223334'
        assert (slot1.get('GPUs') == 3
            and slot1.get('TotalSlotGPUs') == 3
            and slot1.get('AssignedGPUs') == expect_assign1
            and slot2.get('GPUs') == 3
            and slot2.get('TotalSlotGPUs') == 3
            and slot2.get('AssignedGPUs') == expect_assign2)




from typing import Optional


class SecurityContext():
    """
    Alterations to the default configuration used when performing
    authentication between with an HTCondor daemon over the network.
    """

    def __init__(self, token: Optional[str] = None):
        self._preferred_token = token

    def _getPreferredToken(self):
        return self._preferred_token

    def _setPreferredToken(self, token: Optional[str] = None):
        self._preferred_token = token

    def _delPreferredToken(self):
        del self._preferred_token

    preferred_token = property(
        _getPreferredToken,
        _setPreferredToken,
        _delPreferredToken,
        """The token (by value) to try when using this security context."""
    )



# Create the base logger
import logging as _logging
_logger = _logging.getLogger(__name__)
_logger.setLevel(_logging.DEBUG)

import htcondor2

# Set the TMPDIR
from pathlib import Path as _Path
from os import environ as _environ
from os import name as _os_name
from tempfile import gettempdir as _gettempdir
from time import time as _time
TMP_DIR = _Path(_environ.get("_CONDOR_TMPDIR", _gettempdir())) / "htcondor_cli" / str(_time())

# Dict of common command-line options, flags go in the "args" tuple,
# otherwise, use kwargs from:
# https://docs.python.org/3/library/argparse.html#the-add-argument-method
GLOBAL_OPTIONS = {
    "verbose": {
        "args": ("-v", "--verbose"),
        "action": "count",
        "default": 0,
        "help": "Increase verbosity of output (can be specified multiple times)",
    },
    "version": {
        "args": ("-V", "--version"),
        "action": "version",
        "version": htcondor2.version(),
        "help": "Print the HTCondor version and exit"
    },
    "quiet": {
        "args": ("-q", "--quiet"),
        "action": "count",
        "default": 0,
        "help": "Decrease verbosity of output (can be specified multiple times)",
    },
}

# Global objects

# Must be consistent with job status definitions in
# src/condor_includes/proc.h
JobStatus = [
    "NONE",
    "IDLE",
    "RUNNING",
    "REMOVED",
    "COMPLETED",
    "HELD",
    "TRANSFERRING_OUTPUT",
    "SUSPENDED",
    "JOB_STATUS_MAX"
]

# Must be consistent with dag status definitions in
# src/condor_utils/dagman_utils.h
DagStatus = [
    "OK",
    "ERROR",
    "FAILED-NODE",
    "ABORT-ON-SIGNAL",
    "REMOVED",
    "CYCLE",
    "HALTED"
]

class MutualExclusionArgs():
    """Class for lower argparse arguments to declare mutually exclusive arguments"""
    def __init__(self, arguments: dict, title: str = None, desc: str = None) -> None:
        self.args = arguments
        self.title = title
        self.desc = desc

    def make_group(self, parser):
        """Make a mutual exclusion argument group within the provided argparser"""
        group = parser.add_argument_group(self.title, self.desc)
        return group.add_mutually_exclusive_group()

    def __getitem__(self, key):
        return self.args[key]

    def __setitem__(self, key, value):
        self.args[key] = value

    def __len__(self):
        return len(self.args)

    def __iter__(self):
        return iter(self.args)

    def keys(self):
        return self.args.keys()

    def values(self):
        return self.args.values()

    def items(self):
        return self.args.items()

# Import all of the noun classes (do this section last) and then
# create an OrderedDict of nouns, mapping the name to be used on the
# command line to the name of the class containing the noun's verbs.
from collections import OrderedDict as _OrderedDict
from htcondor_cli.dagman import DAG
from htcondor_cli.job import Job
from htcondor_cli.job_set import JobSet
from htcondor_cli.eventlog import EventLog
from htcondor_cli.credential import Credential
from htcondor_cli.server import Server
from htcondor_cli.access_point import AccessPoint
from htcondor_cli.central_manager import CentralManager
from htcondor_cli.ocu import OCU
NOUNS = _OrderedDict()
NOUNS["dag"] = DAG
NOUNS["job"] = Job
NOUNS["jobset"] = JobSet
NOUNS["eventlog"] = EventLog
NOUNS["credential"] = Credential
NOUNS["server"] = Server
NOUNS["ap"] = AccessPoint
NOUNS["cm"] = CentralManager
NOUNS["ocu"] = OCU

# annex needs fcntl which does not exist on windows
if _os_name != 'nt':
    from htcondor_cli.annex import Annex
    NOUNS["annex"] = Annex



#!/usr/bin/env python3
"""
Generate GitHub Actions matrix from nmi-build-platforms CSV file.
Reads nmi_tools/nmi-build-platforms and outputs the matrix configuration
for .github/workflows/ci-build-test.yml as JSON.
"""

import sys
import json


def parse_csv_file(csv_path):
    """Parse the nmi-build-platforms CSV file and return matrix entries."""
    matrix_entries = []
    
    with open(csv_path, 'r') as f:
        for line in f:
            line = line.strip()
            
            # Skip comments and blank lines
            if not line or line.startswith('#'):
                continue
            
            # Parse CSV line
            parts = [p.strip() for p in line.split(',')]
            if len(parts) < 3:
                continue
            
            image_raw = parts[0]
            abbrev = parts[1]
            vm = parts[2]
            
            # Skip batlab entries
            if vm.lower() == 'batlab':
                continue
            
            # Extract container image from docker:// prefix
            if image_raw.startswith('docker://'):
                container = image_raw[9:]  # Remove 'docker://' prefix
                
                matrix_entries.append({
                    'img': vm,
                    'abbrev': abbrev,
                    'container': container
                })
    
    return matrix_entries


def main():
    csv_path = 'nmi_tools/nmi-build-platforms'
    
    try:
        matrix_entries = parse_csv_file(csv_path)
        
        if not matrix_entries:
            print("Warning: No valid entries found in CSV file", file=sys.stderr)
            return 1
        
        # Output as JSON for GitHub Actions
        print(json.dumps({'instance': matrix_entries}))
        
        return 0
        
    except FileNotFoundError:
        print(f"Error: CSV file not found at {csv_path}", file=sys.stderr)
        return 1
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


if __name__ == '__main__':
    sys.exit(main())



import os
import sys
from pathlib import Path

from docutils import nodes, utils
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import custom_ext_parser, make_headerlink_node, warn

KNOB_DEFS = []

# Special default macro grouping names
# i.e. filename -> pretty name
DEFAULT_GROUP_PRETTY_NAME = {
    "dagman": "DAGMan",
    "ssh_to_job": "SSH to Job",
}

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def macro_def_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    global KNOB_DEFS

    default_group = Path(inliner.document.settings.env.docname).name

    # Create a new linkable target using the macro name
    knob, grouping = custom_ext_parser(text)
    targetid = knob
    targetnode = nodes.target('', knob, ids=[targetid], classes=["macro-def"])

    if grouping != "":
        # Specific macro grouping defined inline i.e. :macro-def:`foo[Group One]`
        pass
    elif default_group in DEFAULT_GROUP_PRETTY_NAME:
        grouping = DEFAULT_GROUP_PRETTY_NAME[default_group]
    else:
        grouping = default_group.replace("_", " ").title()

    if knob in KNOB_DEFS:
        docname = inliner.document.settings.env.docname
        warn(f"{docname}:{lineno} | '{knob}' configuration knob already defined!")
        textnode = nodes.Text(knob, " ")
        return [textnode], []
    else:
        KNOB_DEFS.append(knob)

    # Automatically include an index entry for macro definitions
    indexnode = addnodes.index()
    indexnode['entries'] = process_index_entry(f"pair: {knob}; {grouping} Configuration Options", targetid)
    set_role_source_info(inliner, lineno, indexnode)
    headerlink_node = make_headerlink_node(knob, options)

    return [indexnode, targetnode, headerlink_node], []

def setup(app):
    app.add_role("macro-def", macro_def_role)



#!/usr/bin/env pytest

#
# Test the JobEventLog code.
#
# To make sure we test each event type, run x_write_joblog.exe.  Then
# instantiate a JobEventLog and make sure that the JobEvents it returns
# are what we expect them to be.
#

import pytest
import logging

import time
import pickle

import htcondor2 as htcondor

from ornithology import (
	action,
	run_command
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# To make sure we test each event type, run x_write_joblog.exe.
@action
def logfile(pytestconfig, test_dir):
	logFilePath = test_dir / "local.pybind-jel.log"
	exeFilePath = pytestconfig.invocation_dir/ "x_write_joblog.exe"
	completedProcess = run_command([exeFilePath, logFilePath.as_posix()])
	assert(completedProcess.returncode == 0)
	return logFilePath.as_posix()


@action
def synthetic(logfile):
	new = ""
	jel = htcondor.JobEventLog(logfile)
	for event in jel.events(stop_after=0):
		new = new + str(event) + "...\n"
	return new


@action
def original(logfile):
	old = ""
	log = open(logfile, "r")
	for line in log:
		old += line
	return old


standardHost = "<128.105.165.12:32779>"
standardUsage = "Usr 1 02:03:04, Sys 5 06:07:08"

expectedEvents = [
	( 0, {
		"UserNotes": "User info", "LogNotes": "DAGMan info",
		"SubmitHost": standardHost,
	} ),
	( 1, { "ExecuteHost": standardHost } ),
	( 2, { "ExecuteErrorType": 1 } ),
	( 3, {
		"RunRemoteUsage": standardUsage, "RunLocalUsage": standardUsage,
		"SentBytes": 11,
	} ),
	( 4, {
		"RunRemoteUsage": standardUsage, "RunLocalUsage": standardUsage, "SentBytes": 1,
		"TerminatedAndRequeued": True,
		"TerminatedNormally": True, "ReceivedBytes": 2,
		"Checkpointed": False, "Reason": "It misbehaved!",
	} ),
	( 5, {
		"TotalLocalUsage": standardUsage, "TotalRemoteUsage": standardUsage,
		"RunRemoteUsage": standardUsage, "RunLocalUsage": standardUsage,
		"TotalReceivedBytes": 800000, "SentBytes": 400000,
		"TerminatedBySignal": 9, "TotalSentBytes": 900000,
		"CoreFile": "badfilecore", "ReceivedBytes": 200000,
		"TerminatedNormally": False,
	} ),
	( 6, { "Size": 128, "ResidentSetSize": 129, "MemoryUsage": 131, "ProportionalSetSize": 130 } ),
	( 7, { "SentBytes": 4096, "Message": "shadow message", "ReceivedBytes": 4096 } ),
	( 8, { "Info": "Lorem ipsum dolor sit amet, in duo prima aeque principes euismod." } ),
	( 9, { "Reason": "cause I said so!" } ),
	(10, { "NumberOfPIDs": 99 } ),
	(11, { } ),
	(12, { "HoldReasonCode": 404, "HoldReasonSubCode": 255, "HoldReason": 'CauseWeCan' } ),
	(13, { "Reason": "MessinWithYou" } ),
	(14, { "ExecuteHost": standardHost, "Node": 49 } ),
	(15, { "SentBytes": 400000, "RequestMemory": 44, "RequestPets": 1,
		"Memory": 55, "TerminatedNormally": False, "ReturnValue": -1,
		      "PetsUsage": 0.5, "ReceivedBytes": 200000,
		      "TotalRemoteUsage": standardUsage, "Node": 44, "Pets": 1,
		      "CoreFile": "badfilecore", "TotalSentBytes": 900000,
		      "TotalLocalUsage": standardUsage, "TerminatedBySignal": 9,
		      "RunLocalUsage": standardUsage, "RunRemoteUsage": standardUsage,
		      "AssignedPets": "Spot", "MemoryUsage": 33,
		      "TotalReceivedBytes": 800000 } ),
	(16, { "TerminatedBySignal": 9, "TerminatedNormally": False } ),
	#17: { "RestartableJM": True, "RMContact": "ResourceManager", "JMContact": "JobManager" },
	#18: { "Reason": "Cause it could" },
	#19: { "RMContact": "ResourceUp" },
	#20: { "RMContact": "ResourceDown" },
	(21, { "ErrorMsg": "this is the write test error string",
		   "Daemon": "<write job log test>",
		   "ExecuteHost": standardHost } ),
	(22, { "EventDescription": "Job disconnected, attempting to reconnect",
	      "StartdName": "ThatMachine", "DisconnectReason": "TL;DR",
	      "StartdAddr": "<128.105.165.12:32780>" }),
	(23, { "EventDescription": "Job reconnected", "StartdName": "ThatMachine",
	      "StarterAddr": "<128.105.165.12:32780>", "StartdAddr": standardHost } ),
	(24, { "EventDescription": "Job reconnect impossible: rescheduling job",
	      "Reason": "The're just not into you", "StartdName": "ThatMachine" } ),
	(25, { "GridResource": "Resource Name" } ),
	(26, { "GridResource": "Resource Name" } ),
	(27, { "GridResource": "Resource Name", "GridJobId": "100.1" } ),
	(28, { "JobStatus": 2, "BILLReal" : 66.66,
	      "BillString": "lorem ipsum dolor", "BILLBool": True, "BILLInt": 1000 } ),
	(29, { }),
	(30, { }),
	(31, { "EventHead": "Job is performing stage-in of input files" } ),
	(32, { "EventHead": "Job is performing stage-out of output files" } ),
	(33, { "Attribute": "PrivateAttr", "Value": "1" } ),
	(34, { "SkipEventLogNotes": "DAGMan info" } ),
	(35, { "SubmitHost": standardHost } ),
	(36, { "Completion": 1, "NextProcId": 100, "NextRow": 10 } ),
	(37, { "Reason": "Hang on a second", "HoldCode": 24, "PauseCode": 42 } ),
	(38, { "Reason": "just messin wit' ya" } ),
	# 39: ULOG_NONE,
	(40, { "Type": htcondor.FileTransferEventType.IN_QUEUED } ),
	(40, { "Type": htcondor.FileTransferEventType.IN_STARTED } ),
	(40, { "Type": htcondor.FileTransferEventType.IN_FINISHED } ),
	(40, { "Type": htcondor.FileTransferEventType.OUT_QUEUED } ),
	(40, { "Type": htcondor.FileTransferEventType.OUT_STARTED } ),
	(40, { "Type": htcondor.FileTransferEventType.OUT_FINISHED } ),
	# 41: ULOG_RESERVE_SPACE,
	# 42: ULOG_RELEASE_SPACE,
	# 43: ULOG_FILE_COMPLETE,
	# 44: ULOG_FILE_USED,
	# 45: ULOG_FILE_REMOVED,
	# 46 : ULOG_DATAFLOW_JOB_SKIPPED,
	# The default str() conversion of htcondor.CommonFilesEventType.Constant
	# starts with "CommonFilesEventType.", which is not helpful.
	(47, { "Type": "TransferStarted" } ),
	(47, { "Type": "TransferFinished" } ),
	(47, { "Type": "WaitStarted" } ),
]

boringAttrs = ("Subproc", "Proc", "Cluster", "EventTime", "EventTypeNumber", "MyType", "timestamp", "type")


def compareEvent(event, count):
	if event.type != expectedEvents[count][0]:
		print("Event {0} has wrong type".format(count))
		return False

	if event.cluster != 14:
		print("Event {0} has wrong cluster ID".format(count))
		return False

	if event.proc != 55:
		print("Event {0} has wrong proc ID".format(count))
		return False

	if time.time() - event.timestamp > 30:
		print("Event {0}'s timestamp is from more than thirty seconds ago.".format(count))
		return False

	expectedEvent = expectedEvents[count]
	if expectedEvent is not None:
		expectedType = expectedEvent[0]
		expectedAttrs = expectedEvent[1]

		for (attr, value) in expectedAttrs.items():
			if attr not in event:
				print("Event {0}'s {1} is missing".format(count, attr))
				return False
			if event[attr] != value:
				print("Event {0}'s {1} was '{2}', not '{3}'".format(count, attr, event[attr], value))
				return False
		d = set(event.keys()) - set(expectedAttrs.keys()) - set(boringAttrs)
		if d:
			print("Found extra attributes in event {0}: {1}".format(count, ", ".join(sorted(d))))
			return False
	else:
		# To help add new events to the test.
		print("Unknown event {0}:".format(count))
		for attr in event:
			if attr in boringAttrs:
				continue
			print("{0} = '{1}'".format(attr, event[attr]))
		return False

	return True

@action
def negative_cluster_log(test_dir):
	log = test_dir / "negative_cluster.log"
	with open(log, "w") as f:
		f.write("""008 (-01.-01.-01) 01/28 06:28:38 Global JobLog: ctime=1738067318 id=0.3192890.1738067309.170263.12.1738067318.512105 sequence=12 size=0 events=0 offset=23622537909 event_off=0 max_rotation=1 creator_name=<>
...
""")
	return log.as_posix()

class TestJobEventLog:

	# Instantiate a JobEventLog and make sure that the JobEvents it returns
	# are what we expect them to be.
	def test_correct_events_read(self, logfile):
		count = 0
		jel = htcondor.JobEventLog(logfile)
		for event in jel.events(stop_after=0):
			assert(compareEvent(event, count))
			count += 1

		assert(count == len(expectedEvents))


	# To check that __exit__() closes the log, we need to leave at least one
	# event in the log for the log to not return.
	def test_enter_and_exit(self, logfile):
		with htcondor.JobEventLog(logfile) as jel:
			for i in range(0, 30):
				event = next(jel)
		try:
			event = next(jel)
			assert(False)
		except StopIteration as si:
			pass


	# To test __str__(), use it to reconstruct the original log string and
	# then compare to the value of the log file.
	def test_string_conversion(self, synthetic, original):
		assert(synthetic == original)


	def test_close(self, logfile):
		with htcondor.JobEventLog(logfile) as jel:
			e = next(jel)
			jel.close()
			try:
				e = next(jel)
				assert(False)
			except StopIteration as si:
				pass


	def test_pickle(self, logfile):
		p = None
		with htcondor.JobEventLog(logfile) as jel:
			e = next(jel)
			assert(e["UserNotes"] == "User info")
			p = pickle.dumps(jel)

		# Make sure we throw away the pickled JobEventLog.
		del jel

		new_jel = pickle.loads(p)
		e = next(new_jel)

		assert(e["ExecuteHost"] == "<128.105.165.12:32779>")

	def test_non_job_event_handling(self, negative_cluster_log):
		with htcondor.JobEventLog(negative_cluster_log) as jel:
			for event in jel.events(stop_after=0):
				assert event.cluster == -1



import os
from htc_helpers import *

# Global table of documented ClassAd attributes (:classad-attribute-def:)
ATTRIBUTE_FILES = {}
# Table of Specified ClassAd attr type to file
AD_TYPE_FILES = {
    "ACCOUNTING" : "accounting-classad-attributes.html",
    "ADDED_COLLECTOR" : "classad-attributes-added-by-collector.html",
    "COLLECTOR" : "collector-classad-attributes.html",
    "DAEMONCORE" : "daemon-core-statistics-attributes.html",
    "DEFRAG" : "defrag-classad-attributes.html",
    "GRID" : "grid-classad-attributes.html",
    "JOB" : "job-classad-attributes.html",
    "MACHINE" : "machine-classad-attributes.html",
    "MASTER" : "daemon-master-classad-attributes.html",
    "NEGOTIATOR" : "negotiator-classad-attributes.html",
    "SCHEDD" : "scheduler-classad-attributes.html",
    "SUBMITTER" : "submitter-classad-attributes.html",
}

def map_attrs(dir: str):
    """Read ClassAd attribute definition files for documented attributes"""
    attrs = {}
    files_dir = os.path.join(dir, "classad-attributes")
    for ad_file in os.listdir(files_dir):
        if ad_file[-4:] != ".rst":
            continue
        path = os.path.join(files_dir, ad_file)
        with open(path, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                for attr in get_all_defined_role("classad-attribute-def", line):
                    html_file = ad_file.replace(".rst", ".html")
                    if attr in attrs:
                        attrs[attr].append(html_file)
                    else:
                        attrs.update({attr : [html_file]})
    return attrs

def descope_classad(attr: str) -> str:
    """Removed ClassAd attribute scope from attribute"""
    if attr.lower().startswith("my."):
        attr = attr[3:]
    elif attr.lower().startswith("target."):
        attr = attr[7:]
    return attr

def classad_attr_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """Parse inline ad-attr role"""
    docname = inliner.document.settings.env.docname
    root_dir = get_rel_path_to_root_dir(inliner)[:-1]
    full_name, attr_info = custom_ext_parser(text)
    details = extra_info_parser(attr_info) if attr_info != "" else None

    attr_name = descope_classad(full_name)
    ad_type = details.get("TYPE", "").upper() if details is not None else None
    attr_index = details.get("INDEX", "") if details is not None else ""
    filename = AD_TYPE_FILES.get(ad_type) if ad_type in AD_TYPE_FILES else None

    if attr_name not in ATTRIBUTE_FILES:
        warn(f"{docname}:{lineno} | ClassAd Attribute '{attr_name}' not defined in any ClassAd Documentation files")
        filename = "classad-types.html"
    elif filename is None:
        filename = ATTRIBUTE_FILES[attr_name][0]
        if len(ATTRIBUTE_FILES[attr_name]) > 1:
            warn(f"{docname}:{lineno} | ClassAd Attribute '{attr_name}' is defined in multiple files. Defaulting to {filename}")

    ref_link = f"href=\"{root_dir}/classad-attributes/{filename}#{attr_name}\""
    return make_ref_and_index_nodes(name, full_name, attr_index, ref_link,
                                    rawtext, inliner, lineno, options, attr_name)

def setup(app):
    """Setup ad-attr role"""
    global ATTRIBUTE_FILES
    # Create mapping of ClassAd attribute to source file
    ATTRIBUTE_FILES = map_attrs(app.srcdir)
    app.add_role("ad-attr", classad_attr_role)




import sys
import enum


class Value(enum.IntEnum):
    """
    An enumeration of special ClassAd values.

    .. attribute:: Undefined

    .. attribute:: Error
    """

    Error        = 1<<0
    Undefined    = 1<<1



#!/usr/bin/env pytest

import htcondor2

import logging
from getpass import getuser

from ornithology import *


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def local_dir(test_dir):
    return test_dir / "condor"


@action
def the_condor(local_dir):
    with Condor(
        local_dir=local_dir,
        raw_config="""
            # Enable OAuth2.
            SEC_CREDENTIAL_DIRECTORY_OAUTH      = $(LOCAL_DIR)/oauth.d
            DAEMON_LIST                         = $(DAEMON_LIST),CREDD
            TRUSTED_VAULT_HOSTS                 = trusted.org, alsotrusted.edu
        """,
    ) as the_condor:
        yield the_condor


@action
def trusted_token_file(test_dir):
    token_file = test_dir / "the_trusted_token"
    token_file.write_text('{"vault_token": "XXX", "vault_url": "https://trusted.org"}')
    return token_file


@action
def untrusted_token_file(test_dir):
    token_file = test_dir / "the_untrusted_token"
    token_file.write_text('{"vault_token": "XXX", "vault_url": "https://untrusted.org"}')
    return token_file


class TestVaultCredential:

    def test_trusted_credential(self, local_dir, the_condor, trusted_token_file):
        rv = the_condor.run_command(
            ['condor_store_cred', 'add-oauth', '-s', 'vault1',
             '-i', str(trusted_token_file)],
        )
        logger.info(rv.stdout)
        logger.info(rv.stderr)
        assert rv.returncode == 0

        top_file = local_dir / "oauth.d" / getuser() / "vault1.top"
        assert trusted_token_file.read_text() == top_file.read_text()

    def test_untrusted_credential(self, local_dir, the_condor, untrusted_token_file):
        rv = the_condor.run_command(
            ['condor_store_cred', 'add-oauth', '-s', 'vault2',
             '-i', str(untrusted_token_file)],
        )
        logger.info(rv.stdout)
        logger.info(rv.stderr)
        assert rv.returncode != 0



#!/usr/bin/env pytest

import time

from ornithology import (
    action,
    ClusterState,
    Condor,
    format_script,
    write_file,
)

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


expected_output = '''Starting counting from 0
0
1
2
3
4
Checkpointing because I feel like it...
Starting counting from 5
5
6
7
8
9
Checkpointing because I feel like it...
Starting counting from 10
... finished computing.
'''


TEST_CASES = {
    # You really want this case first, so we can call condor_reschedule
    # while it's running and save ourselves a bunch of time.
    "true": {
        "config": {
            "CLAIM_WORKLIFE":           0,
            "CHECK_REACTIVATE_AFTER_CHECKPOINT":    "TRUE",
        },
        "expected": True,
    },
    "true_but_useless": {
        "config": {
            "CLAIM_WORKLIFE":           999999,
            "CHECK_REACTIVATE_AFTER_CHECKPOINT":    "TRUE",
        },
        "expected": False,
    },
    "false": {
        "config": {
            "CLAIM_WORKLIFE":           0,
            "CHECK_REACTIVATE_AFTER_CHECKPOINT":    "FALSE",
        },
        "expected": False,
    },
    "unset": {
        "config": {
            "CLAIM_WORKLIFE":           0,
            # This can't be completely unset because it inherits from places.
            "CHECK_REACTIVATE_AFTER_CHECKPOINT":    "",
        },
        "expected": False,
    }
}


@action
def the_condors(test_dir):
    the_condors = {}

    for case in TEST_CASES.keys():
        local_dir = test_dir / f"condor.{case}"

        the_condor = Condor(
            local_dir=local_dir,
            config=TEST_CASES[case]['config'],
        )
        the_condor._start()
        the_condors[case] = the_condor

    yield the_condors

    for the_condor in the_condors.values():
        the_condor._cleanup()


@action
def the_test_script(test_dir):
    path = test_dir / "self-checkpoint.py"
    contents = format_script(
    '''
        #!/usr/bin/python3

        import sys
        import time
        from pathlib import Path


        def SaveState():
            global i
            Path("saved-state").write_text(str(i))


        def VacateJob():
            print("Saving state on vacate...")
            SaveState()
            sys.exit(85)


        def SpontaneousCheckpoint():
            print("Checkpointing because I feel like it...")
            SaveState()
            sys.exit(85)


        i = 0
        saved_state_file = Path("saved-state")
        if saved_state_file.exists():
            i = int(saved_state_file.read_text())

        print(f"Starting counting from {i}")
        while i != 10:
            time.sleep(1)

            print(i)

            i = i + 1

            if i % 5 == 0:
                SpontaneousCheckpoint()

        print( "... finished computing.")
        Path("output").touch()
    '''
    )
    write_file(path, contents)

    return path


@action
def the_jobs(test_dir, the_condors, the_test_script):
    description = {
        "universe":                 "vanilla",
        "executable":               the_test_script.as_posix(),
        "checkpoint_exit_code":     85,
        "when_to_transfer_output":  "ON_EXIT",
        "transfer_executable":      True,
        "should_transfer_files":    True,
        "request_cpus":             1,
        "request_memory":           1,
    }

    the_jobs = {}
    for case, the_condor in the_condors.items():
        the_job = the_condor.submit(
            description={
                ** description,
                "log":              (test_dir / f"{case}.log").as_posix(),
                "output":           (test_dir / f"{case}.out").as_posix(),
            },
            count=1,
        )
        the_jobs[case] = the_job

    return the_jobs


@action(params={case: case for case in TEST_CASES})
def the_job_pair(request, the_jobs):
    return (request.param, the_jobs[request.param])


@action
def the_case(the_job_pair):
    return the_job_pair[0]


@action
def the_job(the_job_pair):
    return the_job_pair[1]


@action
def the_completed_job(the_case, the_condors, the_job):
    # For unknown reasons(s), the schedd doesn't actually reschedule jobs
    # that were evicted immediately, so in some cases we have to kick it.

    we_should_wait = the_case == "true"
    while we_should_wait:

        print("Waiting for job to become idle...")
        assert the_job.wait(
            timeout=60,
            condition=ClusterState.all_idle,
            fail_condition=ClusterState.any_held,
        )

        print("... rescheduling...")
        the_condors[the_case].run_command(['condor_reschedule'])

        print("... waiting for job to start running...")
        assert the_job.wait(
            timeout=60,
            condition=ClusterState.all_running,
            fail_condition=ClusterState.any_held,
        )

        print("... waiting to see if the job becomes idle again...");
        we_should_wait = the_job.wait(
            timeout=60,
            condition=ClusterState.all_idle,
            fail_condition=ClusterState.any_terminal,
        )


    assert the_job.wait(
        timeout=300,
        condition=ClusterState.all_complete,
        fail_condition=ClusterState.any_held,
    )

    return the_job


class TestRestartIsReactivate():

    def test_job_completed_successfully(self, test_dir, the_case, the_completed_job):
        # I'd love to tunnel this filename though the fixture system...
        actual_output_path = test_dir / f"{the_case}.out"
        actual_output = actual_output_path.read_text()
        assert actual_output == expected_output

        # I'd love to tunnel this filename though the fixture system...
        actual_log_path = test_dir / f"{the_case}.log"
        actual_log = actual_log_path.read_text()
        assert TEST_CASES[the_case]['expected'] == ('Rescheduling self-checkpoint job after checkpoint upload because reactivating the claim would have failed.' in actual_log)



from flask import Flask, render_template, request
from blueprints.landing import landing_bp
from blueprints.overview import overview_bp
import os
import utils

################################################
# Main application setup
################################################

# Create Flask app
app = Flask(__name__)

################################################
# Configuration parameters (all settable via environment variables)
################################################

# CE_DASHBOARD_CACHE_DIRECTORY is the directory where the cache files will be stored.
# This directory should be writable by the web server user. If it does not exist, it will be created.
# Default to '/tmp/ce_dashboard_cache'.
app.config['CE_DASHBOARD_CACHE_DIRECTORY'] = os.getenv('CE_DASHBOARD_CACHE_DIRECTORY', '/tmp/ce_dashboard_cache')  

# CE_DEFAULT_CE_DOMAIN is the default CE domain to use if the host parameter is not fully qualified.
app.config['CE_DASHBOARD_DEFAULT_CE_DOMAIN'] = os.getenv('CE_DASHBOARD_DEFAULT_CE_DOMAIN', 'svc.opensciencegrid.org')


# CE_DASHBOARD_SERVER_CACHE_MINUTES is the number of minutes to cache data on this server.
# Default to 20 minutes.
app.config['CE_DASHBOARD_SERVER_CACHE_MINUTES'] = int(os.getenv('CE_DASHBOARD_SERVER_CACHE_MINUTES', '20'))

# CE Dashboard browser cache minutes is the number of minutes to cache data in the browser.
# Default to 30 minutes.
app.config['CE_DASHBOARD_BROWSER_CACHE_MINUTES'] = int(os.getenv('CE_DASHBOARD_BROWSER_CACHE_MINUTES', '30'))

# CE_DASHBOARD_METRICSD_URL is the URL of the MetricsD server to query for CE Dashboard data.
# Default to 'https://display.ospool.osg-htc.org'.
app.config['CE_DASHBOARD_METRICSD_URL'] = os.getenv('CE_DASHBOARD_METRICSD_URL', 'https://display.ospool.osg-htc.org')


################################################
# Initialize my utils and any other modules
################################################

utils.set_flask_app(app)

################################################
# Create routes in common for multiple blueprints
################################################

common_linkmap = {
    'Overview': 'overview.html',
    'Contributed': 'contributed.html',
}

@app.route('/fullscreen.html')
def handle_fullscreen():
    return render_template('fullscreen.html', linkmap=common_linkmap)  

@app.route('/edit.html')
def handle_edit():
    return render_template('edit.html', linkmap=common_linkmap)

# Register blueprints
app.register_blueprint(landing_bp)
app.register_blueprint(overview_bp)

if __name__ == '__main__':
    app.run(debug=True)



#!/usr/bin/env pytest

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

import time
from ornithology import *


@action
def the_attr():
    return f"attr_{int(time.time() % 60)}"


@action
def the_value():
    return f"{int((time.time() % 3600)/60)}"


@action
def the_condor(test_dir, the_attr, the_value):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "DAEMON_LIST":                  "TEST_DC_STD_FUNCTIOND COLLECTOR",
            "TEST_DC_STD_FUNCTIOND":        "$(LIBEXEC)/test_dc_std_functiond",
            "TEST_DC_STD_FUNCTIOND_ARGS":   f"{the_attr} {the_value}",
            # Ornithology needs this, for some reason.
            "TEST_DC_STD_FUNCTIOND_LOG":    f"$(LOG)/TestDcStdFunctionLog",
            "DC_DAEMON_LIST":               "+ TEST_DC_STD_FUNCTIOND",
        },
    ) as condor:
        yield condor


@action
def the_reply(the_condor, the_attr):
    the_log = the_condor._get_daemon_log('TEST_DC_STD_FUNCTIOND')
    for line in the_log.open().read():
        if line.message.startswith('DaemonCore: command socket at '):
            [prefix, sinful] = line.message.split('DaemonCore: command socket at ')
            result = the_condor.run_command(
                ['condor_status', '-af', the_attr, '-direct', sinful]
            )
            assert result.returncode == 0
            return result.stdout
    assert False


class TestDCStdFunction:

    def test_replied_correctly(self, the_reply, the_value):
        assert the_reply == the_value
